# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for workflow_util tool."""
import collections
import copy
import mock
import pickle
import sys
import unittest

from pinball.config.pinball_config import PinballConfig
from pinball.master.thrift_lib.ttypes import GroupRequest
from pinball.master.thrift_lib.ttypes import GroupResponse
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import ModifyResponse
from pinball.master.thrift_lib.ttypes import Query
from pinball.master.thrift_lib.ttypes import QueryAndOwnRequest
from pinball.master.thrift_lib.ttypes import QueryAndOwnResponse
from pinball.master.thrift_lib.ttypes import QueryRequest
from pinball.master.thrift_lib.ttypes import QueryResponse
from pinball.master.thrift_lib.ttypes import Token
from pinball.scheduler.schedule import WorkflowSchedule
from pinball.tools.workflow_util import Abort
from pinball.tools.workflow_util import Cleanup
from pinball.tools.workflow_util import Disable
from pinball.tools.workflow_util import Drain
from pinball.tools.workflow_util import Enable
from pinball.tools.workflow_util import Exit
from pinball.tools.workflow_util import Pause
from pinball.tools.workflow_util import Poison
from pinball.tools.workflow_util import RebuildCache
from pinball.tools.workflow_util import Redo
from pinball.tools.workflow_util import Reload
from pinball.tools.workflow_util import ReSchedule
from pinball.tools.workflow_util import Retry
from pinball.tools.workflow_util import Resume
from pinball.tools.workflow_util import Start
from pinball.tools.workflow_util import Stop
from pinball.tools.workflow_util import UnAbort
from pinball.tools.workflow_util import UnDrain
from pinball.tools.workflow_util import UnExit
from pinball.tools.workflow_util import UnSchedule
from pinball.ui.data import Status
from pinball.ui.data import WorkflowData
from pinball.ui.data import WorkflowInstanceData
from pinball.workflow.event import Event
from pinball.workflow.job import ShellJob
from pinball.workflow.job_executor import ExecutionRecord
from pinball.workflow.signaller import Signal


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = ['Pawel Garbacki', 'Mao Ye']
__license__ = 'Apache'
__version__ = '2.0'


def _cmp_job_tokens(test, job_token1, job_token2):
    token1 = copy.copy(job_token1)
    token2 = copy.copy(job_token2)
    job1 = pickle.loads(token1.data)
    job2 = pickle.loads(token2.data)
    test.assertEqual(str(job1), str(job2))
    token1.data = None
    token2.data = None
    test.assertEqual(token1, token2)


class StartTestCase(unittest.TestCase):
    @mock.patch('pinball.parser.utils.load_path')
    @mock.patch('pinball.tools.workflow_util._check_workflow_instances')
    def test_start_non_existent(self, check_workflow_instances_mock, load_path_mock):
        Options = collections.namedtuple('args', 'workflow')
        options = Options(workflow='does_not_exist')
        PinballConfig.PARSER_PARAMS = {'key': 'value'}
        command = Start()
        command.prepare(options)
        check_workflow_instances_mock.return_value = True
        config_parser = mock.Mock()

        def load_path(params):
            self.assertEqual(['caller', 'key'], sorted(params.keys()))
            return config_parser
        load_path_mock.return_value = load_path
        config_parser.get_workflow_tokens.return_value = None

        client = mock.Mock()
        output = command.execute(client, None)

        self.assertEqual('workflow does_not_exist not found in %s\n' %
                         str(PinballConfig.PARSER_PARAMS), output)

    @mock.patch('pinball.parser.utils.load_path')
    @mock.patch('pinball.tools.workflow_util._check_workflow_instances')
    def test_start_workflow(self, check_workflow_instance_mock, load_path_mock):
        Options = collections.namedtuple('args', 'workflow')
        options = Options(workflow='some_workflow')
        PinballConfig.PARSER_PARAMS = {'key': 'value'}
        command = Start()
        command.prepare(options)
        check_workflow_instance_mock.return_value = True
        config_parser = mock.Mock()

        def load_path(params):
            self.assertEqual(['caller', 'key'], sorted(params.keys()))
            return config_parser
        load_path_mock.return_value = load_path
        job_token = Token(
            name='/workflow/some_workflow/123/job/waiting/some_job',
            data='some_data')
        event_token = Token(name=('/workflow/some_workflow/123/input/some_job/'
                                  '__WORKFLOW_START__'),
                            data='some_data')
        config_parser.get_workflow_tokens.return_value = [job_token,
                                                          event_token]
        client = mock.Mock()

        output = command.execute(client, None)

        config_parser.get_workflow_tokens.assert_called_once_with(
            'some_workflow')
        self.assertEqual('exported workflow some_workflow instance 123.  '
                         'Its tokens are under '
                         '/workflow/some_workflow/123/', output)


class StopTestCase(unittest.TestCase):
    def test_stop_non_existent(self):
        Options = collections.namedtuple('args', 'workflow, instance, force')
        options = Options(workflow='does_not_exist', instance='123',
                          force=True)
        command = Stop()
        command.prepare(options)

        response = QueryResponse(tokens=[[]])
        client = mock.Mock()
        client.query.return_value = response

        output = command.execute(client, None)

        query = Query(namePrefix='/workflow/does_not_exist/123/')
        query_request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(query_request)
        self.assertEqual('workflow does_not_exist instance 123 not found\n',
                         output)

    def test_stop_workflow(self):
        Options = collections.namedtuple('args', 'workflow, instance, force')
        options = Options(workflow='some_workflow', instance='123', force=True)
        command = Stop()
        command.prepare(options)

        token = Token(version=1234567,
                      name='/workflow/some_workflow/123/job/waiting/some_job',
                      data='some_data')
        query_response = QueryResponse(tokens=[[token]])
        client = mock.Mock()
        client.query.return_value = query_response
        client.modify.return_value = ModifyResponse()

        output = command.execute(client, None)

        query = Query(namePrefix='/workflow/some_workflow/123/')
        query_request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(query_request)
        modify_request = ModifyRequest(deletes=[token])
        client.modify.assert_called_once_with(modify_request)
        self.assertEqual('removed 1 token(s) in 1 tries\n', output)


class PauseTestCase(unittest.TestCase):
    def test_pause_non_existent(self):
        Options = collections.namedtuple('args', 'workflow, instance, force')
        options = Options(workflow='does_not_exist', instance='123',
                          force=True)
        command = Pause()
        command.prepare(options)

        response = QueryResponse(tokens=[[]])
        client = mock.Mock()
        client.query.return_value = response

        output = command.execute(client, None)

        query = Query(namePrefix='/workflow/does_not_exist/123/')
        query_request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(query_request)
        self.assertEqual('workflow does_not_exist instance 123 not found\n',
                         output)

    def test_pause_workflow(self):
        Options = collections.namedtuple('args', 'workflow, instance, force')
        options = Options(workflow='some_workflow', instance='123', force=True)
        command = Pause()
        command.prepare(options)

        token = Token(version=1234567,
                      name='/workflow/some_workflow/123/job/waiting/some_job',
                      data='some_data')
        original_token = copy.copy(token)
        query_response = QueryResponse(tokens=[[token]])
        client = mock.Mock()
        client.query.return_value = query_response
        client.modify.return_value = ModifyResponse()

        output = command.execute(client, None)

        query = Query(namePrefix='/workflow/some_workflow/123/')
        query_request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(query_request)
        self.assertEqual(1, client.modify.call_count)
        modify_request = client.modify.call_args[0][0]
        self.assertEqual(1, len(modify_request.updates))
        owned_token = modify_request.updates[0]
        self.assertEqual(owned_token.name, original_token.name)
        self.assertEqual(owned_token.version, original_token.version)
        self.assertTrue(owned_token.owner.startswith('workflow_util'))
        self.assertEqual(sys.maxint, owned_token.expirationTime)
        self.assertEqual('claimed 1 token(s) in 1 tries\n', output)


class ResumeTestCase(unittest.TestCase):
    def test_resume_non_existent(self):
        Options = collections.namedtuple('args', 'workflow, instance, force')
        options = Options(workflow='does_not_exist', instance='123',
                          force=True)
        command = Resume()
        command.prepare(options)

        response = QueryResponse(tokens=[[]])
        client = mock.Mock()
        client.query.return_value = response

        output = command.execute(client, None)

        query = Query(namePrefix='/workflow/does_not_exist/123/')
        query_request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(query_request)
        self.assertEqual('workflow does_not_exist instance 123 not found\n',
                         output)

    def test_resume_workflow(self):
        Options = collections.namedtuple('args', 'workflow, instance, force')
        options = Options(workflow='some_workflow', instance='123', force=True)
        command = Resume()
        command.prepare(options)

        token = Token(version=1234567,
                      name='/workflow/some_workflow/123/job/waiting/some_job',
                      data='some_data',
                      owner='some_owner',
                      expirationTime=sys.maxint)
        query_response = QueryResponse(tokens=[[token]])
        client = mock.Mock()
        client.query.return_value = query_response
        client.modify.return_value = ModifyResponse()

        output = command.execute(client, None)

        query = Query(namePrefix='/workflow/some_workflow/123/')
        query_request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(query_request)
        updated_token = copy.copy(token)
        updated_token.owner = None
        updated_token.expirationTime = None
        modify_request = ModifyRequest(updates=[updated_token])
        client.modify.assert_called_once_with(modify_request)
        self.assertEqual('released ownership of 1 token(s)\n', output)


class RetryTestCase(unittest.TestCase):
    def setUp(self):
        self._signal_token = Token(
            name='/workflow/some_workflow/123/__SIGNAL__/ARCHIVE')

        succeeded_job = ShellJob()
        succeeded_job.history = [ExecutionRecord(exit_code=0)]
        self._succeeded_job_token = Token(
            name='/workflow/some_workflow/123/job/waiting/succeeded',
            data=pickle.dumps(succeeded_job))

        failed_job = ShellJob()
        failed_job.history = [ExecutionRecord(exit_code=1)]
        self._failed_job_token = Token(
            name='/workflow/some_workflow/123/job/waiting/failed',
            priority=10,
            data=pickle.dumps(failed_job))

    @mock.patch('pinball.tools.workflow_util._check_workflow_instances')
    def test_retry_non_existent(self, check_workflow_instances_mock):
        Options = collections.namedtuple('args', 'workflow, instance, force')
        options = Options(workflow='does_not_exist', instance='123',
                          force=True)
        command = Retry()
        command.prepare(options)

        response = GroupResponse()
        client = mock.Mock()
        client.group.return_value = response

        check_workflow_instances_mock.return_value = True

        store = mock.Mock()
        store.read_archived_tokens.return_value = []

        output = command.execute(client, store)

        group_request = GroupRequest(
            namePrefix='/workflow/does_not_exist/123/job/', groupSuffix='/')
        client.group.assert_called_once_with(group_request)
        self.assertEqual('workflow does_not_exist instance 123 not found\n',
                         output)

    @mock.patch('pinball.tools.workflow_util._check_workflow_instances')
    def test_retry_active_workflow(self, check_workflow_instances_mock):
        Options = collections.namedtuple('args', 'workflow, instance, force')
        options = Options(workflow='some_workflow', instance='123', force=True)
        command = Retry()
        command.prepare(options)
        check_workflow_instances_mock.return_value = True

        client = mock.Mock()

        group_response = GroupResponse(
            counts={'/workflow/some_workflow/123/job': 2})
        client.group.return_value = group_response

        query_response = QueryResponse(tokens=[
            [self._signal_token],
            [self._succeeded_job_token, self._failed_job_token]])
        client.query.return_value = query_response

        modify_response = ModifyResponse(updates=[self._failed_job_token])
        client.modify.return_value = modify_response

        def side_effect(request):
            self.assertEqual([self._signal_token, self._failed_job_token],
                             request.deletes)

            failed_job_token = copy.copy(self._failed_job_token)
            failed_job_token.name = (
                '/workflow/some_workflow/123/job/runnable/failed')
            self.assertEqual(1, len(request.updates))
            _cmp_job_tokens(self, failed_job_token, request.updates[0])
            # self.assertEqual(failed_job_token, request.updates[0])

        client.modify.side_effect = side_effect

        output = command.execute(client, None)

        group_request = GroupRequest(
            namePrefix='/workflow/some_workflow/123/job/', groupSuffix='/')
        client.group.assert_called_once_with(group_request)

        signal_query = Query(
            namePrefix='/workflow/some_workflow/123/__SIGNAL__/ARCHIVE')
        job_query = Query(
            namePrefix='/workflow/some_workflow/123/job/waiting/')
        query_request = QueryRequest(queries=[signal_query, job_query])
        client.query.assert_called_once_with(query_request)

        self.assertEqual('retried 1 job(s) in workflow some_workflow instance '
                         '123\n', output)

    @mock.patch('pinball.tools.workflow_util.'
                'get_unique_workflow_instance')
    @mock.patch('pinball.tools.workflow_util._check_workflow_instances')
    def test_retry_archived_workflow(self, check_workflow_instances_mock,
                                     get_unique_workflow_instance_mock):
        Options = collections.namedtuple('args', 'workflow, instance, force')
        options = Options(workflow='some_workflow', instance='123', force=True)
        command = Retry()
        command.prepare(options)

        client = mock.Mock()

        group_response = GroupResponse()
        client.group.return_value = group_response

        check_workflow_instances_mock.return_value = True
        get_unique_workflow_instance_mock.return_value = '321'

        store = mock.Mock()
        store.read_archived_tokens.return_value = [self._signal_token,
                                                   self._succeeded_job_token,
                                                   self._failed_job_token]

        succeeded_job_token = copy.copy(self._succeeded_job_token)
        succeeded_job_token.name = (
            '/workflow/some_workflow/123/job/waiting/succeeded')
        failed_job_token = copy.copy(self._failed_job_token)
        failed_job_token.name = (
            '/workflow/some_workflow/123/job/runnable/failed')
        modify_response = ModifyResponse(updates=[succeeded_job_token,
                                                  failed_job_token])
        client.modify.return_value = modify_response

        output = command.execute(client, store)

        group_request = GroupRequest(
            namePrefix='/workflow/some_workflow/123/job/',
            groupSuffix='/')
        client.group.assert_called_once_with(group_request)

        store.read_archived_tokens.assert_called_once_with(
            name_prefix='/workflow/some_workflow/123/')

        succeeded_job_token.name = (
            '/workflow/some_workflow/321/job/waiting/succeeded')
        failed_job_token = copy.copy(self._failed_job_token)
        failed_job_token.name = (
            '/workflow/some_workflow/321/job/runnable/failed')
        modify_request = ModifyRequest(updates=[succeeded_job_token,
                                                failed_job_token])
        client.modify.assert_called_once_with(modify_request)

        self.assertEqual('retried workflow some_workflow instance 123.  Its '
                         'tokens are under /workflow/some_workflow/321/\n',
                         output)


class RedoTestCase(unittest.TestCase):
    @mock.patch('pinball.tools.workflow_util.get_unique_name')
    @mock.patch('time.time')
    def test_redo_non_existent(self, time_mock, get_unique_name_mock):
        get_unique_name_mock.return_value = 'some_owner'
        time_mock.return_value = 10
        Options = collections.namedtuple('args', 'workflow, instance, jobs, '
                                         'execution, force')
        options = Options(workflow='does_not_exist', instance='123',
                          jobs='some_job', execution='1', force=True)
        command = Redo()
        command.prepare(options)

        response = QueryAndOwnResponse()
        client = mock.Mock()
        client.query_and_own.return_value = response

        output = command.execute(client, None)

        query = Query(
            namePrefix='/workflow/does_not_exist/123/job/waiting/some_job')
        request = QueryAndOwnRequest(owner='some_owner',
                                     expirationTime=(10 + 60), query=query)
        client.query_and_own.assert_called_once_with(request)
        self.assertEqual('workflow must be running, the job must be finished '
                         'and it cannot be runnable', output)

    @staticmethod
    def _get_waiting_job_token():
        job = ShellJob()
        execution_record_with_events = ExecutionRecord()
        execution_record_with_events.events = [Event()]
        job.history = [ExecutionRecord(), execution_record_with_events]
        return Token(
            name='/workflow/some_workflow/123/job/waiting/some_job',
            data=pickle.dumps(job))

    @staticmethod
    def _get_runnable_job_token(waiting_job_token):
        job = pickle.loads(waiting_job_token.data)
        job.events = job.history[1].events
        return Token(name='/workflow/some_workflow/123/job/runnable/some_job',
                     data=pickle.dumps(job))

    @mock.patch('pinball.tools.workflow_util.get_unique_name')
    @mock.patch('time.time')
    def test_redo(self, time_mock, get_unique_name_mock):
        get_unique_name_mock.return_value = 'some_owner'
        time_mock.return_value = 10
        Options = collections.namedtuple('args', 'workflow, instance, jobs, '
                                         'execution, force')
        options = Options(workflow='some_workflow', instance='123',
                          jobs='some_job', execution='1', force=True)
        command = Redo()
        command.prepare(options)

        client = mock.Mock()
        waiting_job = RedoTestCase._get_waiting_job_token()
        query_and_own_response = QueryAndOwnResponse(tokens=[waiting_job])
        client.query_and_own.return_value = query_and_own_response

        runnable_job = RedoTestCase._get_runnable_job_token(waiting_job)
        modify_response = ModifyResponse(updates=[runnable_job])
        client.modify.return_value = modify_response

        output = command.execute(client, None)

        query = Query(
            namePrefix='/workflow/some_workflow/123/job/waiting/some_job')
        query_and_own_request = QueryAndOwnRequest(
            owner='some_owner', expirationTime=(10 + 60), query=query)
        client.query_and_own.assert_called_once_with(query_and_own_request)

        modify_request = ModifyRequest(updates=[runnable_job],
                                       deletes=[waiting_job])
        client.modify.assert_called_once_with(modify_request)

        self.assertEqual('redoing execution 1 of job some_job in workflow '
                         'some_workflow instance 123\n', output)


class PoisonTestCase(unittest.TestCase):
    def setUp(self):
        self._signal_token = Token(
            name='/workflow/some_workflow/123/__SIGNAL__/ARCHIVE')

        parent_job = ShellJob(name='parent', inputs=['__WORKFLOW_START__'],
                              outputs=['child'])
        self._parent_job_token = Token(
            name='/workflow/some_workflow/123/job/waiting/parent',
            data=pickle.dumps(parent_job))

        child_job = ShellJob(name='child', inputs=['parent'])
        self._child_job_token = Token(
            name='/workflow/some_workflow/123/job/waiting/child',
            data=pickle.dumps(child_job))

        PinballConfig.PARSER_PARAMS = {'key': 'value'}

    def test_poison_non_existent(self):
        Options = collections.namedtuple(
            'args', 'workflow, instance, jobs, force')
        options = Options(workflow='does_not_exist', instance='123',
                          jobs='parent', force=True)
        command = Poison()
        command.prepare(options)
        client = mock.Mock()

        group_response = GroupResponse(counts={})
        client.group.return_value = group_response

        store = mock.Mock()
        store.read_archived_tokens.return_value = []

        output = command.execute(client, store)

        group_request = GroupRequest(
            namePrefix='/workflow/does_not_exist/123/job/',
            groupSuffix='/')
        client.group.assert_called_once_with(group_request)

        store.read_archived_tokens.assert_called_with(
            name_prefix='/workflow/does_not_exist/123/')
        self.assertEqual('workflow does_not_exist instance 123 not found\n',
                         output)

    def test_poison_from_store(self):
        Options = collections.namedtuple(
            'args', 'workflow, instance, jobs, force')
        options = Options(workflow='some_workflow', instance='123',
                          jobs='parent', force=True)
        command = Poison()
        command.prepare(options)
        client = mock.Mock()

        store = mock.Mock()
        store.read_archived_tokens.return_value = [self._parent_job_token,
                                                   self._child_job_token]

        group_response = GroupResponse(counts={})
        client.group.return_value = group_response

        modify_response = ModifyResponse(updates=[self._parent_job_token,
                                                  self._child_job_token])
        client.modify.return_value = modify_response

        def side_effect(request):
            # Expect two job tokens and one event token.
            self.assertEqual(3, len(request.updates))
            self.assertIsNone(request.deletes)

        client.modify.side_effect = side_effect

        output = command.execute(client, store)

        group_request = GroupRequest(
            namePrefix='/workflow/some_workflow/123/job/',
            groupSuffix='/')
        client.group.assert_called_once_with(group_request)

        store.read_archived_tokens.assert_called_with(
            name_prefix='/workflow/some_workflow/123/')
        self.assertTrue(output.startswith("poisoned workflow some_workflow "
                                          "roots ['parent']."))

    @mock.patch('pinball.tools.workflow_util.Analyzer')
    def test_poison_from_client(self, analyzer_mock):
        Options = collections.namedtuple(
            'args', 'workflow, instance, jobs, force')
        options = Options(workflow='some_workflow', instance='123',
                          jobs='parent', force=True)
        command = Poison()
        command.prepare(options)
        client = mock.Mock()

        analyzer = mock.Mock()
        analyzer_mock.from_client.return_value = analyzer
        analyzer.get_tokens.return_value = [self._parent_job_token,
                                            self._child_job_token]

        event_token = Token(
            version=1234567,
            name='/workflow/some_workflow/123/input/some_job/some_input',
            data='some_data')
        analyzer.get_new_event_tokens.return_value = [event_token]

        group_response = GroupResponse(
            counts={'/workflow/some_workflow/123/job/': 1})
        client.group.return_value = group_response

        query_response = QueryResponse(tokens=[[self._signal_token]])
        client = mock.Mock()
        client.query.return_value = query_response

        modify_response = ModifyResponse(updates=[event_token])
        client.modify.return_value = modify_response

        output = command.execute(client, None)

        group_request = GroupRequest(
            namePrefix='/workflow/some_workflow/123/job/',
            groupSuffix='/')
        client.group.assert_called_once_with(group_request)
        query = Query(
            namePrefix='/workflow/some_workflow/123/__SIGNAL__/ARCHIVE')
        query_request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(query_request)
        modify_request = ModifyRequest(updates=[event_token],
                                       deletes=[self._signal_token])
        client.modify.assert_called_once_with(modify_request)

        self.assertEqual(1, analyzer.get_tokens.call_count)
        self.assertEqual(1, analyzer.get_new_event_tokens.call_count)

        self.assertEqual("poisoned workflow some_workflow instance 123 roots "
                         "['parent']\n", output)

    @mock.patch('pinball.tools.workflow_util.Analyzer')
    def test_poison_from_config(self, analyzer_mock):
        Options = collections.namedtuple(
            'args', 'workflow, instance, jobs, force')
        options = Options(workflow='some_workflow', instance=None,
                          jobs='parent', force=True)
        command = Poison()
        command.prepare(options)

        analyzer = mock.Mock()
        analyzer_mock.from_parser_params.return_value = analyzer
        analyzer.get_tokens.return_value = [self._parent_job_token,
                                            self._child_job_token]

        client = mock.Mock()
        modify_response = ModifyResponse(updates=[self._parent_job_token,
                                                  self._child_job_token])
        client.modify.return_value = modify_response
        store = mock.Mock()

        output = command.execute(client, store)

        analyzer_mock.from_parser_params.assert_called_once_with('some_workflow')
        self.assertEqual(2, analyzer.get_tokens.call_count)
        self.assertEqual(1, analyzer.clear_job_histories.call_count)
        analyzer.poison.assert_called_once_with(['parent'])
        self.assertEqual(1, analyzer.change_instance.call_count)
        modify_request = ModifyRequest(updates=[self._parent_job_token,
                                                self._child_job_token])
        client.modify.assert_called_once_with(modify_request)
        self.assertTrue(output.startswith("poisoned workflow some_workflow "
                                          "roots ['parent']."))


class ModifySignalTestCase(unittest.TestCase):
    def _init(self, command, action):
        self._command = command
        self._action = action

    def _run_test(self, is_present, expect_set, expect_remove, expected_output,
                  signaller_mock):
        Options = collections.namedtuple('args', 'workflow, instance, force')
        options = Options(workflow='some_workflow', instance='123', force=True)
        self._command.prepare(options)

        signaller_mock.is_signal_present.return_value = is_present

        output = self._command.execute(mock.Mock(), None)
        self.assertEqual(expected_output, output)

        if expect_set:
            signaller_mock.set_action.assert_called_with(self._action)

        if expect_remove:
            signaller_mock.remove_action.assert_called_with(self._action)


class DrainTestCase(ModifySignalTestCase):
    def setUp(self):
        self._init(Drain(), Signal.DRAIN)

    @mock.patch('pinball.tools.workflow_util.Signaller')
    def test_set_present(self, signaller_mock):
        signaller = mock.Mock()
        signaller_mock.return_value = signaller
        expected_output = ('DRAIN has been already set.  Not changing '
                           'anything this time\n')
        self._run_test(True, False, False, expected_output, signaller)

    @mock.patch('pinball.tools.workflow_util.Signaller')
    def test_set_missing(self, signaller_mock):
        signaller = mock.Mock()
        signaller_mock.return_value = signaller
        expected_output = ('set DRAIN.  Its token is '
                           '/workflow/some_workflow/123/__SIGNAL__/DRAIN\n')
        self._run_test(False, True, False, expected_output, signaller)


class UnDrainTestCase(ModifySignalTestCase):
    def setUp(self):
        self._init(UnDrain(), Signal.DRAIN)

    @mock.patch('pinball.tools.workflow_util.Signaller')
    def test_remove_present(self, signaller_mock):
        signaller = mock.Mock()
        signaller_mock.return_value = signaller
        expected_output = ('removed DRAIN from '
                           '/workflow/some_workflow/123/__SIGNAL__/DRAIN\n')
        self._run_test(True, False, True, expected_output, signaller)

    @mock.patch('pinball.tools.workflow_util.Signaller')
    def test_remove_missing(self, signaller_mock):
        signaller = mock.Mock()
        signaller_mock.return_value = signaller
        expected_output = ('DRAIN has been already removed.  Not changing '
                           'anything this time\n')
        self._run_test(False, False, False, expected_output, signaller)


class AbortTestCase(ModifySignalTestCase):
    def setUp(self):
        self._init(Abort(), Signal.ABORT)

    @mock.patch('pinball.tools.workflow_util.Signaller')
    def test_set_present(self, signaller_mock):
        signaller = mock.Mock()
        signaller_mock.return_value = signaller
        expected_output = ('ABORT has been already set.  Not changing '
                           'anything this time\n')
        self._run_test(True, False, False, expected_output, signaller)

    @mock.patch('pinball.tools.workflow_util.Signaller')
    def test_set_missing(self, signaller_mock):
        signaller = mock.Mock()
        signaller_mock.return_value = signaller
        expected_output = ('set ABORT.  Its token is '
                           '/workflow/some_workflow/123/__SIGNAL__/ABORT\n')
        self._run_test(False, True, False, expected_output, signaller)


class UnAbortTestCase(ModifySignalTestCase):
    def setUp(self):
        self._init(UnAbort(), Signal.ABORT)

    @mock.patch('pinball.tools.workflow_util.Signaller')
    def test_remove_present(self, signaller_mock):
        signaller = mock.Mock()
        signaller_mock.return_value = signaller
        expected_output = ('removed ABORT from '
                           '/workflow/some_workflow/123/__SIGNAL__/ABORT\n')
        self._run_test(True, False, True, expected_output, signaller)

    @mock.patch('pinball.tools.workflow_util.Signaller')
    def test_remove_missing(self, signaller_mock):
        signaller = mock.Mock()
        signaller_mock.return_value = signaller
        expected_output = ('ABORT has been already removed.  Not changing '
                           'anything this time\n')
        self._run_test(False, False, False, expected_output, signaller)


class ExitTestCase(ModifySignalTestCase):
    def setUp(self):
        self._init(Exit(), Signal.EXIT)

    @mock.patch('pinball.tools.workflow_util.Signaller')
    def test_set_present(self, signaller_mock):
        signaller = mock.Mock()
        signaller_mock.return_value = signaller
        expected_output = ('EXIT has been already set.  Not changing '
                           'anything this time\n')
        self._run_test(True, False, False, expected_output, signaller)

    @mock.patch('pinball.tools.workflow_util.Signaller')
    def test_set_missing(self, signaller_mock):
        signaller = mock.Mock()
        signaller_mock.return_value = signaller
        expected_output = ('set EXIT.  Its token is '
                           '/workflow/some_workflow/123/__SIGNAL__/EXIT\n')
        self._run_test(False, True, False, expected_output, signaller)


class UnExitTestCase(ModifySignalTestCase):
    def setUp(self):
        self._init(UnExit(), Signal.EXIT)

    @mock.patch('pinball.tools.workflow_util.Signaller')
    def test_remove_present(self, signaller_mock):
        signaller = mock.Mock()
        signaller_mock.return_value = signaller
        expected_output = ('removed EXIT from '
                           '/workflow/some_workflow/123/__SIGNAL__/EXIT\n')
        self._run_test(True, False, True, expected_output, signaller)

    @mock.patch('pinball.tools.workflow_util.Signaller')
    def test_remove_missing(self, signaller_mock):
        signaller = mock.Mock()
        signaller_mock.return_value = signaller
        expected_output = ('EXIT has been already removed.  Not changing '
                           'anything this time\n')
        self._run_test(False, False, False, expected_output, signaller)


class ReScheduleTestCase(unittest.TestCase):
    @mock.patch('pinball.parser.utils.load_path')
    def test_reschdule_non_existent(self, load_path_mock):
        Options = collections.namedtuple('args',
                                         'workflow,force')
        options = Options(workflow='does_not_exist', force=True)
        PinballConfig.PARSER_PARAMS = {'key': 'value'}
        command = ReSchedule()
        command.prepare(options)

        config_parser = mock.Mock()

        def load_path(params):
            self.assertEqual(['caller', 'key'], sorted(params.keys()))
            return config_parser
        load_path_mock.return_value = load_path
        config_parser.get_workflow_names.return_value = []

        client = mock.Mock()
        output = command.execute(client, None)

        self.assertEqual('workflow does_not_exist not found\n', output)

    @mock.patch('pinball.parser.utils.load_path')
    def test_reschedule_workflow(self, load_path_mock):
        Options = collections.namedtuple('args',
                                         'workflow,force')
        options = Options(workflow='some_workflow', force=True)
        PinballConfig.PARSER_PARAMS = {'key': 'value'}
        command = ReSchedule()
        command.prepare(options)

        config_parser = mock.Mock()

        def load_path(params):
            self.assertEqual(['caller', 'key'], sorted(params.keys()))
            return config_parser
        load_path_mock.return_value = load_path
        config_parser.get_workflow_names.return_value = ['some_workflow']
        schedule = WorkflowSchedule(next_run_time=10,
                                    recurrence_seconds=10,
                                    workflow='some_workflow',
                                    emails=None)
        old_token = Token(version=1234567,
                          name='/schedule/workflow/some_workflow',
                          data=pickle.dumps(schedule),
                          owner='some_owner',
                          expirationTime=sys.maxint)
        new_token = copy.copy(old_token)
        schedule.recurrence_seconds = 100
        new_token.data = pickle.dumps(schedule)
        config_parser.get_schedule_token.return_value = new_token
        query_response = QueryResponse(tokens=[[old_token]])
        client = mock.Mock()
        client.query.return_value = query_response
        client.modify.return_value = ModifyResponse()

        output = command.execute(client, None)

        query = Query(namePrefix='/schedule/workflow/some_workflow')
        query_request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(query_request)
        modify_request = ModifyRequest(updates=[new_token])
        client.modify.assert_called_once_with(modify_request)
        self.assertEqual("rescheduled workflows ['some_workflow'].  Their "
                         "schedule tokens are "
                         "['/schedule/workflow/some_workflow']\n", output)

    @mock.patch('pinball.parser.utils.load_path')
    def test_reschedule_workflows(self, load_path_mock):
        Options = collections.namedtuple('args',
                                         'workflow, force')
        options = Options(workflow=None, force=True)
        PinballConfig.PARSER_PARAMS = {'key': 'value'}
        command = ReSchedule()
        command.prepare(options)

        config_parser = mock.Mock()

        def load_path(params):
            self.assertEqual(['caller', 'key'], sorted(params.keys()))
            return config_parser
        load_path_mock.return_value = load_path

        config_parser.get_workflow_names.return_value = ['some_workflow',
                                                         'some_other_workflow']

        schedule = WorkflowSchedule(next_run_time=10,
                                    recurrence_seconds=10,
                                    workflow='some_workflow',
                                    emails=None)
        old_token = Token(version=1234567,
                          name='/schedule/workflow/some_workflow',
                          data=pickle.dumps(schedule),
                          owner='some_owner',
                          expirationTime=sys.maxint)
        new_token = copy.copy(old_token)
        schedule.recurrence_seconds = 100
        new_token.data = pickle.dumps(schedule)

        schedule.workflow = 'some_other_workflow'
        other_old_token = Token(version=12345678,
                                name='/schedule/workflow/some_other_workflow',
                                data=pickle.dumps(schedule),
                                owner='some_owner',
                                expirationTime=sys.maxint)
        other_new_token = copy.copy(other_old_token)
        other_new_token.data = pickle.dumps(schedule)

        def get_schedule_token(workflow):
            if workflow == 'some_workflow':
                return new_token
            elif workflow == 'some_other_workflow':
                return other_new_token
        config_parser.get_schedule_token = get_schedule_token

        query_response = QueryResponse(tokens=[[old_token], [other_old_token]])
        client = mock.Mock()
        client.query.return_value = query_response
        client.modify.return_value = ModifyResponse()

        output = command.execute(client, None)

        query = Query(namePrefix='/schedule/workflow/some_workflow')
        other_query = Query(
            namePrefix='/schedule/workflow/some_other_workflow')
        query_request = QueryRequest(queries=[query, other_query])
        client.query.assert_called_once_with(query_request)
        # other_new_token is not in the updates because the schedule of
        # some_other_workflow doesn't change.
        modify_request = ModifyRequest(updates=[new_token])
        client.modify.assert_called_once_with(modify_request)
        self.assertEqual("rescheduled workflows ['some_workflow'].  Their "
                         "schedule tokens are "
                         "['/schedule/workflow/some_workflow']\n", output)


class UnScheduleTestCase(unittest.TestCase):
    def test_unschdule_non_existent(self):
        Options = collections.namedtuple('args', 'workflow,force')
        options = Options(workflow='does_not_exist', force=True)
        command = UnSchedule()
        command.prepare(options)

        query_response = QueryResponse(tokens=[[]])
        client = mock.Mock()
        client.query.return_value = query_response

        output = command.execute(client, None)

        query = Query(namePrefix='/schedule/workflow/does_not_exist')
        query_request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(query_request)
        self.assertEqual('schedule for workflow does_not_exist not found\n',
                         output)

    def test_reschedule_workflow(self):
        Options = collections.namedtuple('args', 'workflow,force')
        options = Options(workflow='some_workflow', force=True)
        command = UnSchedule()
        command.prepare(options)

        token = Token(version=1234567,
                      name='/schedule/workflow/some_workflow',
                      data='some_data',
                      owner='some_owner',
                      expirationTime=sys.maxint)
        query_response = QueryResponse(tokens=[[token]])
        client = mock.Mock()
        client.query.return_value = query_response
        client.modify.return_value = ModifyResponse()

        output = command.execute(client, None)

        query = Query(namePrefix='/schedule/workflow/some_workflow')
        query_request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(query_request)
        modify_request = ModifyRequest(deletes=[token])
        client.modify.assert_called_once_with(modify_request)
        self.assertEqual('removed schedule for workflow some_workflow\n',
                         output)


class ReloadTestCase(unittest.TestCase):
    def setUp(self):
        not_owned_job = ShellJob(max_attempts=5)
        not_owned_job.history = [ExecutionRecord(exit_code=0)]
        self._not_owned_job_token = Token(
            name='/workflow/some_workflow/123/job/runnable/not_owned',
            priority=10,
            data=pickle.dumps(not_owned_job))

        not_owned_job.max_attempts = 6
        self._new_not_owned_job_token = Token(
            name='/workflow/some_workflow/123/job/runnable/not_owned',
            priority=11,
            data=pickle.dumps(not_owned_job))

        not_owned_job.history = []
        self._not_owned_job_def_token = Token(
            name='/workflow/some_workflow/123/job/waiting/not_owned',
            priority=11,
            data=pickle.dumps(not_owned_job))

        owned_job = ShellJob()
        self._owned_job_token = Token(
            name='/workflow/some_workflow/123/job/runnable/owned',
            owner='some_owner',
            expirationTime=1000,
            data=pickle.dumps(owned_job))

        self._owned_job_def_token = Token(
            name='/workflow/some_workflow/123/job/waiting/owned',
            data=pickle.dumps(owned_job))

        PinballConfig.PARSER_PARAMS = {'key': 'value'}

    @mock.patch('pinball.parser.utils.load_path')
    def test_reload_non_existent(self, load_path_mock):
        Options = collections.namedtuple(
            'args', 'workflow, instance, jobs, force')
        options = Options(workflow='does_not_exist',
                          instance='123',
                          jobs=None,
                          force=True)
        PinballConfig.PARSER_PARAMS = {'key': 'value'}
        command = Reload()
        command.prepare(options)

        config_parser = mock.Mock()

        def load_path(params):
            self.assertEqual(['caller', 'key'], sorted(params.keys()))
            return config_parser
        load_path_mock.return_value = load_path
        config_parser.get_workflow_names.return_value = []

        client = mock.Mock()
        output = command.execute(client, None)

        self.assertEqual('workflow does_not_exist not found in %s\n' %
                         str(PinballConfig.PARSER_PARAMS), output)

    @mock.patch('pinball.tools.workflow_util.time')
    @mock.patch('pinball.parser.utils.load_path')
    def test_reload_owned_job(self, load_path_mock, time):
        Options = collections.namedtuple(
            'args', 'workflow, instance, jobs, force')
        options = Options(workflow='some_workflow',
                          instance='123',
                          jobs='owned',
                          force=True)
        PinballConfig.PARSER_PARAMS = {'key': 'value'}
        command = Reload()
        command.prepare(options)

        config_parser = mock.Mock()

        def load_path(params):
            self.assertEqual(['caller', 'key'], sorted(params.keys()))
            return config_parser
        load_path_mock.return_value = load_path
        config_parser.get_workflow_names.return_value = ['some_workflow']

        time.time.return_value = 10

        client = mock.Mock()
        client.query_and_own.return_value = QueryAndOwnResponse(tokens=[])

        output = command.execute(client, None)

        runnable_query = Query(
            namePrefix='/workflow/some_workflow/123/job/runnable/owned',
            maxTokens=1)
        runnable_request = QueryAndOwnRequest(owner='workflow_util',
                                              expirationTime=(10 + 5 * 60),
                                              query=runnable_query)

        waiting_query = Query(
            namePrefix='/workflow/some_workflow/123/job/waiting/owned',
            maxTokens=1)
        waiting_request = QueryAndOwnRequest(owner='workflow_util',
                                             expirationTime=(10 + 5 * 60),
                                             query=waiting_query)
        self.assertEqual([mock.call(runnable_request),
                          mock.call(waiting_request)],
                         client.query_and_own.call_args_list)

        self.assertEqual('job owned in workflow some_workflow instance 123 '
                         'either not found or already owned', output)

    @staticmethod
    def _check_modify_request(expected_request):
        """Verify a modify request.

        We cannot simply compare the exact content of the tokens passed in the
        modify request as pickling could have distorted it.

        Args:
            expected_request: The request to compare with.
        Returns:
            Function with the same interface as TokenMasterService.modify()
        """
        def check_request(request):
            assert not expected_request.deletes and not request.deletes
            assert expected_request.updates and not request.updates
            assert len(expected_request.updates) == len(request.updates)
            for i in range(0, len(expected_request.updates)):
                assert (str(expected_request.updates[i]) ==
                        str(request.updates[i]))
        return check_request

    @mock.patch('pinball.tools.workflow_util.time')
    @mock.patch('pinball.parser.utils.load_path')
    def test_reload_not_owned_job(self, load_path_mock, time):
        Options = collections.namedtuple(
            'args', 'workflow, instance, jobs, force')
        options = Options(workflow='some_workflow',
                          instance='123',
                          jobs='not_owned',
                          force=True)
        PinballConfig.PARSER_PARAMS = {'key': 'value'}
        command = Reload()
        command.prepare(options)

        config_parser = mock.Mock()

        def load_path(params):
            self.assertEqual(['caller', 'key'], sorted(params.keys()))
            return config_parser
        load_path_mock.return_value = load_path
        config_parser.get_workflow_names.return_value = ['some_workflow']

        event_token = Token(name=('/workflow/some_workflow/123/input/'
                                  'not_owned_job/__WORKFLOW_START__'),
                            data='some_data')
        config_parser.get_workflow_tokens.return_value = [
            self._not_owned_job_def_token,
            self._owned_job_def_token,
            event_token]

        time.time.return_value = 10

        client = mock.Mock()

        not_owned_job_token = copy.copy(self._not_owned_job_token)
        not_owned_job_token.owner = 'workflow_util'
        not_owned_job_token.expirationTime = 10 + 5 * 60
        client.query_and_own.return_value = QueryAndOwnResponse(
            tokens=[not_owned_job_token])

        client.modify.return_value = ModifyResponse(
            updates=[self._new_not_owned_job_token])

        output = command.execute(client, None)

        config_parser.get_workflow_tokens.assert_called_once_with(
            'some_workflow')

        runnable_query = Query(
            namePrefix='/workflow/some_workflow/123/job/runnable/not_owned',
            maxTokens=1)
        runnable_request = QueryAndOwnRequest(owner='workflow_util',
                                              expirationTime=(10 + 5 * 60),
                                              query=runnable_query)
        client.query_and_own.assert_called_once_with(runnable_request)

        new_not_owned_job_token = copy.copy(self._new_not_owned_job_token)
        new_not_owned_job_token.name = ('/workflow/some_workflow/123/job/'
                                        'runnable/not_owned')
        client.modify.side_effect = ReloadTestCase._check_modify_request(
            ModifyRequest(updates=[new_not_owned_job_token], deletes=None))

        self.assertEqual("reloaded jobs ['not_owned'] in workflow "
                         "some_workflow instance 123", output)

    @mock.patch('pinball.tools.workflow_util.time')
    @mock.patch('pinball.parser.utils.load_path')
    def test_reload_all_jobs(self, load_path_mock, time):
        Options = collections.namedtuple(
            'args', 'workflow, instance, jobs, force')
        options = Options(workflow='some_workflow',
                          instance='123',
                          jobs=None,
                          force=True)
        PinballConfig.PARSER_PARAMS = {'key': 'value'}
        command = Reload()
        command.prepare(options)

        config_parser = mock.Mock()

        def load_path(params):
            self.assertEqual(['caller', 'key'], sorted(params.keys()))
            return config_parser
        load_path_mock.return_value = load_path
        config_parser.get_workflow_names.return_value = ['some_workflow']

        time.time.return_value = 10

        client = mock.Mock()
        client.group.return_value = GroupResponse(counts={
            '/workflow/some_workflow/123/job/runnable': 1,
            '/workflow/some_workflow/123/job/waiting': 1})

        not_owned_job_token = copy.copy(self._not_owned_job_token)
        not_owned_job_token.owner = 'workflow_util'
        not_owned_job_token.expirationTime = 10 + 5 * 60
        client.query_and_own.return_value = QueryAndOwnResponse(
            tokens=[not_owned_job_token])

        client.modify.return_value = ModifyResponse(
            updates=[self._not_owned_job_token])

        output = command.execute(client, None)

        group_request = GroupRequest(
            namePrefix='/workflow/some_workflow/123/job/',
            groupSuffix='/')
        client.group.assert_called_once_with(group_request)

        query = Query(namePrefix='/workflow/some_workflow/123/job/')
        query_and_own_request = QueryAndOwnRequest(
            owner='workflow_util',
            expirationTime=(10 + 5 * 60),
            query=query)
        client.query_and_own.assert_called_once_with(query_and_own_request)

        modify_request = ModifyRequest(updates=[self._not_owned_job_token])
        client.modify.assert_called_once_with(modify_request)

        self.assertEqual('only 1 out of 2 job tokens in workflow '
                         'some_workflow instance 123 could be claimed', output)


class AlterTestCase(unittest.TestCase):
    def _init(self, command, altered_parent_disabled, altered_child_disabled):
        self._command = command
        self._altered_parent_disabled = altered_parent_disabled
        self._altered_child_disabled = altered_child_disabled

    def _alter_non_existent(self, expected_output):
        Options = collections.namedtuple(
            'args', 'workflow, instance, jobs, force')
        options = Options(workflow='does_not_exist', instance='123',
                          jobs='some_job', force=True)
        self._command.prepare(options)

        client = mock.Mock()
        client.query.return_value = QueryResponse(tokens=[[]])
        store = mock.Mock()
        output = self._command.execute(client, store)

        query = Query(namePrefix='/workflow/does_not_exist/123/job/')
        query_request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(query_request)
        self.assertEqual(expected_output, output)

    def _check_modify_request(self, request):
        self.assertEqual(2, len(request.updates))
        parent_token = request.updates[0]
        self.assertEqual('/workflow/some_workflow/123/job/waiting/parent',
                         parent_token.name)
        parent_job = pickle.loads(parent_token.data)
        self.assertEqual(self._altered_parent_disabled, parent_job.disabled)
        child_token = request.updates[1]
        self.assertEqual('/workflow/some_workflow/123/job/waiting/child',
                         child_token.name)
        child_job = pickle.loads(child_token.data)
        self.assertEqual(self._altered_child_disabled, child_job.disabled)

    def _alter_jobs(self, expected_output):
        Options = collections.namedtuple(
            'args', 'workflow, instance, jobs, force')
        options = Options(workflow='some_workflow', instance='123',
                          jobs='parent,child', force=True)
        self._command.prepare(options)

        parent_job = ShellJob(name='parent', inputs=['__WORKFLOW_START__'],
                              outputs=['child'])
        parent_job.disabled = True
        parent_job_token = Token(
            name='/workflow/some_workflow/123/job/waiting/parent',
            data=pickle.dumps(parent_job))

        altered_parent_job_token = copy.copy(parent_job_token)
        parent_job.disabled = self._altered_parent_disabled
        altered_parent_job_token.data = pickle.dumps(parent_job)

        child_job = ShellJob(name='child', inputs=['parent'])
        child_job_token = Token(
            name='/workflow/some_workflow/123/job/waiting/child',
            data=pickle.dumps(child_job))

        altered_child_job_token = copy.copy(child_job_token)
        child_job.disabled = self._altered_child_disabled
        altered_child_job_token.data = pickle.dumps(child_job)

        client = mock.Mock()
        client.query.return_value = QueryResponse(tokens=[[parent_job_token,
                                                           child_job_token]])
        client.modify.return_value = ModifyResponse(
            updates=[altered_parent_job_token, altered_child_job_token])
        client.modify.side_effect = self._check_modify_request

        store = mock.Mock()
        output = self._command.execute(client, store)

        query = Query(namePrefix='/workflow/some_workflow/123/job/')
        query_request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(query_request)

        self.assertEqual(1, client.modify.call_count)

        self.assertEqual(expected_output, output)


class DisableTestCase(AlterTestCase):
    def setUp(self):
        self._init(Disable(), True, True)

    def test_alter_non_existent(self):
        self._alter_non_existent("job(s) ['some_job'] not found in the "
                                 "master.  Note that only jobs of a running "
                                 "workflow can be disabled")

    def test_alter_jobs(self):
        self._alter_jobs('disabled 2 job(s) in 1 tries\n')


class EnableTestCase(AlterTestCase):
    def setUp(self):
        self._init(Enable(), False, False)

    def test_alter_non_existent(self):
        self._alter_non_existent("job(s) ['some_job'] not found in the "
                                 "master.  Note that only jobs of a running "
                                 "workflow can be enabled")

    def test_alter_jobs(self):
        self._alter_jobs('enabled 2 job(s) in 1 tries\n')


class CleanupTestCase(unittest.TestCase):
    def setUp(self):
        # Set PinballConfig to enable s3 log saver
        PinballConfig.S3_LOGS_DIR_PREFIX = 's3n://pinball/tmp/'
        PinballConfig.S3_LOGS_DIR = \
            PinballConfig.S3_LOGS_DIR_PREFIX \
            + PinballConfig.JOB_LOG_PATH_PREFIX

    @mock.patch('pinball.tools.workflow_util.DataBuilder')
    def test_cleanup_non_existent(self, data_builder):
        Options = collections.namedtuple('args', 'age_days,force')
        options = Options(age_days=10, force=True)
        command = Cleanup()
        command.prepare(options)

        builder = mock.Mock()
        data_builder.return_value = builder
        builder.get_workflows.return_value = []

        store = mock.Mock()
        output = command.execute(None, store)

        builder.get_workflows.assert_called_once_with()
        self.assertEqual('no tokens need to be cleaned up\nremoved 0 token(s) '
                         'and 0 directory(ies)\n', output)

    @mock.patch('pinball.tools.workflow_util.delete_s3_directory')
    @mock.patch('pinball.tools.workflow_util.time')
    @mock.patch('pinball.tools.workflow_util.DataBuilder')
    @mock.patch('pinball.tools.workflow_util.shutil')
    def test_cleanup_instances(self, shutil, data_builder, time, delete_s3):
        Options = collections.namedtuple('args', 'age_days,force')
        options = Options(age_days=10, force=True)
        command = Cleanup()

        time.time.return_value = 1373181000  # 2013-07-07 07:10:00 UTC

        command.prepare(options)

        builder = mock.Mock()
        data_builder.return_value = builder

        old_workflow = WorkflowData('old_workflow')
        new_workflow = WorkflowData('new_workflow')
        builder.get_workflows.return_value = [old_workflow, new_workflow]

        old_instance = WorkflowInstanceData('old_workflow', '1234',
                                            status=Status.SUCCESS,
                                            start_time=10,
                                            end_time=20)

        new_instance = WorkflowInstanceData(
            'new_workflow', '12345',
            status=Status.SUCCESS,
            start_time=1373180880,  # 2013-07-07 07:08:00 UTC
            end_time=1373180940)  # 2013-07-07 07:09:00 UTC

        def side_effect(*args, **kwargs):
            if args[0] == old_workflow.workflow:
                return [old_instance]
            elif args[0] == new_workflow.workflow:
                return [new_instance]
            else:
                assert False, 'unknown workflow %s' % args[0]
        builder.get_instances.side_effect = side_effect

        store = mock.Mock()
        token = Token(name='some_token')
        store.read_archived_tokens.return_value = [token]

        output = command.execute(None, store)

        builder.get_workflows.assert_called_once_with()
        self.assertEqual(2, builder.get_instances.call_count)
        builder.get_instances.assert_any_call(old_workflow.workflow)
        builder.get_instances.assert_any_call(new_workflow.workflow)

        store.read_archived_tokens.assert_called_once_with(
            name_prefix='/workflow/old_workflow/1234/')
        store.delete_archived_tokens.assert_called_once_with([token])

        shutil.rmtree.assert_called_once_with(
            '/tmp/pinball_job_logs/old_workflow/1234', ignore_errors=True)

        delete_s3.assert_called_once_with(
            's3n://pinball/tmp/pinball_job_logs/old_workflow/1234')

        self.assertEqual('removed 1 token(s) and 2 directory(ies)\n', output)


class RebuildCacheTestCase(unittest.TestCase):
    @mock.patch('pinball.tools.workflow_util.DataBuilder')
    def test_rebuild_cache(self, data_builder):
        Options = collections.namedtuple('args', 'force')
        options = Options(force=True)
        command = RebuildCache()

        command.prepare(options)

        builder = mock.Mock()
        data_builder.return_value = builder
        store = mock.Mock()
        store.read_cached_data_names.return_value = [
            '/workflow/some_workflow/123',
            '/workflow/some_workflow/1234',
            '/workflow/some_workflow/12345']

        output = command.execute(None, store)

        self.assertEqual(2, store.read_cached_data_names.call_count)
        self.assertEqual(1, store.clear_cached_data.call_count)
        data_builder.assert_called_once_with(store, use_cache=True)
        self.assertEqual(1, builder.get_workflows.call_count)

        self.assertEqual('rebuilt data cache.  It now has 3 data items',
                         output)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Definition of job metadata included in job tokens.

Job object describes job inputs, outputs, and all information required to
execute a job (e.g., a command line of a shell job or class name of a data
job)."""
import abc

from pinball.config.utils import get_log
from pinball.persistence.token_data import TokenData
from pinball.workflow.name import Name


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.workflow.worker')


class Job(TokenData):
    """Parent class for specialized job types."""
    __metaclass__ = abc.ABCMeta

    IS_CONDITION = False

    def __init__(self, name=None, inputs=None, outputs=None, emails=None,
                 max_attempts=1, retry_delay_sec=0, warn_timeout_sec=None,
                 abort_timeout_sec=None):
        self.name = name
        self.inputs = inputs if inputs is not None else []
        self.outputs = outputs if outputs is not None else []
        self.emails = emails if emails is not None else []
        self.max_attempts = max_attempts
        self.retry_delay_sec = retry_delay_sec
        self.warn_timeout_sec = warn_timeout_sec
        self.abort_timeout_sec = abort_timeout_sec
        assert self.max_attempts > 0
        self.disabled = False
        self.history = []
        self.events = []

    @property
    def _COMPATIBILITY_ATTRIBUTES(self):
        return {
            'emails': [],
            'disabled': False,
            'max_attempts': 1,
            'events': [],
            'warn_timeout_sec': None,
            'abort_timeout_sec': None,
            'retry_delay_sec': 0,
        }

    @abc.abstractmethod
    def info(self):
        return

    def retry(self):
        """Decide if the job should be retried.

        Returns:
            True if the job should be retried, otherwise False.
        """
        if not self.history:
            return False
        last_record = self.history[-1]
        current_instance = last_record.instance
        assert last_record.exit_code != 0
        failed_runs = 0
        for record in reversed(self.history):
            if record.instance != current_instance:
                break
            if record.exit_code != 0:
                # There may have been successful runs in the past if we are
                # re-doing an execution.
                failed_runs += 1
            if failed_runs >= self.max_attempts:
                return False
        return True

    def truncate_history(self):
        if self.IS_CONDITION and len(self.history) > self.max_attempts:
            self.history = self.history[-self.max_attempts:]

    def reload(self, new_job):
        """Reload job config from a new config.

        Configuration elements defining the workflow topology (inputs and
        outputs), execution history, or run-time values (events) are not
        modified.

        Args:
            new_job: The new job configuration to update from.
        """
        assert self.__class__ == new_job.__class__
        self.emails = new_job.emails
        self.max_attempts = new_job.max_attempts


class ShellJob(Job):
    """Shell job runs a command when executed."""
    def __init__(self, name=None, inputs=None, outputs=None, emails=None,
                 max_attempts=1, retry_delay_sec=0, warn_timeout_sec=None,
                 abort_timeout_sec=None, command=None, cleanup_template=None):
        super(ShellJob, self).__init__(name, inputs, outputs, emails,
                                       max_attempts, retry_delay_sec,
                                       warn_timeout_sec, abort_timeout_sec)
        self.command = command
        self.cleanup_template = cleanup_template

    @property
    def _COMPATIBILITY_ATTRIBUTES(self):
        result = super(ShellJob, self)._COMPATIBILITY_ATTRIBUTES
        result['cleanup_template'] = None
        return result

    def __str__(self):
        return ('ShellJob(name=%s, inputs=%s, outputs=%s, emails=%s, '
                'max_attempts=%d, retry_delay_sec=%d, warn_timeout_sec=%s, '
                'abort_timeout_sec=%s, disabled=%s, command=%s, '
                'cleanup_template=%s, events=%s, history=%s)' % (
                    self.name,
                    self.inputs,
                    self.outputs,
                    self.emails,
                    self.max_attempts,
                    self.retry_delay_sec,
                    self.warn_timeout_sec,
                    self.abort_timeout_sec,
                    self.disabled,
                    self.command,
                    self.cleanup_template,
                    self.events,
                    self.history))

    def __repr__(self):
        return self.__str__()

    def info(self):
        return 'command=%s' % self.command

    def reload(self, new_job):
        super(ShellJob, self).reload(new_job)
        self.command = new_job.command
        self.cleanup_template = new_job.cleanup_template

    @staticmethod
    def _get_command_attributes(template):
        """Extract attributes from a command string template.

        E.g., for template 'ls %(dir1)s %(dir2)s' the result is
        ['dir1', 'dir2'].

        Args:
            template: The template to extract attributes from.
        Returns:
            The list of named attributes extracted from the template.
        """
        class Extractor:
            """Helper class extracting attributes from a string template.
            """
            def __init__(self):
                self.attributes = set()

            def __getitem__(self, attribute):
                self.attributes.add(attribute)
                return 0

        extractor = Extractor()
        try:
            template % extractor
        except ValueError:
            LOG.exception('failed to customize template %s', template)
        return list(extractor.attributes)

    def _consolidate_event_attributes(self):
        """Consolidate attributes in triggering events.

        Iterate over events in the most recent execution record and combine
        them into one dictionary mapping attribute names to their values.  If
        multiple events contain the same attribute, the return value will be a
        comma separated string of values from all those events.

        Returns:
            Dictionary of consolidated event attribute key-values.
        """
        assert self.history
        last_execution_record = self.history[-1]
        result = {}
        for event in last_execution_record.events:
            for key, value in event.attributes.items():
                new_value = result.get(key)
                if new_value:
                    new_value += ',%s' % value
                else:
                    new_value = value
                result[key] = new_value
        return result

    def customize_command(self):
        """Specialize the command with attribute values extracted from events.

        Returns:
            Job command with parameter values replaced by attributes extracted
            from the triggering events.  If a parameter is not present in the
            event attribute set, it is replaced with an empty string.
        """
        attributes = {}
        command_attributes = ShellJob._get_command_attributes(self.command)
        for attribute in command_attributes:
            attributes[attribute] = ''
        event_attributes = self._consolidate_event_attributes()
        attributes.update(event_attributes)
        try:
            return self.command % attributes
        except ValueError:
            LOG.exception('failed to customize command %s', self.command)
            return self.command


class ShellConditionJob(ShellJob):
    IS_CONDITION = True

    def __init__(self, name=None, outputs=None, emails=None, max_attempts=10,
                 retry_delay_sec=5 * 60, warn_timeout_sec=None,
                 abort_timeout_sec=None, command=None, cleanup_template=None):
        super(ShellConditionJob, self).__init__(
            name=name,
            inputs=[Name.WORKFLOW_START_INPUT],
            outputs=outputs,
            emails=emails,
            max_attempts=max_attempts,
            retry_delay_sec=retry_delay_sec,
            warn_timeout_sec=warn_timeout_sec,
            abort_timeout_sec=abort_timeout_sec,
            command=command,
            cleanup_template=cleanup_template)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import unittest

from pinball_ext.executor import cluster_executor


__author__ = 'Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class ClusterExecutorTest(unittest.TestCase):
    def test_creation(self):
        user_jar_dirs = ['/dir1/jar1/', '/dir2/jar2/']
        user_app_jar = '/dir/jar.jar'
        user_archive = '/dir/file.archive'
        platform = 'local'

        executor = cluster_executor.ClusterExecutor(
            executor_config={
                'USER_LIBJAR_DIRS': ','.join(user_jar_dirs),
                'USER_APPJAR_PATH': user_app_jar,
                'USER_ARCHIVE_PATH': user_archive,
                'PLATFORM': platform,
                'USER': 'test_user'
            }
        )

        self.assertEqual(executor.config.USER_LIBJAR_DIRS, user_jar_dirs)
        self.assertEqual(executor.config.USER_APPJAR_PATH, user_app_jar)
        self.assertEqual(executor.config.USER_ARCHIVE_PATH, user_archive)
        self.assertEqual(executor.config.PLATFORM, platform)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Archiver takes care of archiving workflow instance tokens."""
import time

from pinball.config.utils import get_log
from pinball.master.thrift_lib.ttypes import ArchiveRequest
from pinball.master.thrift_lib.ttypes import Query
from pinball.master.thrift_lib.ttypes import QueryRequest
from pinball.master.thrift_lib.ttypes import TokenMasterException

from pinball.workflow.name import Name
from pinball.workflow.signaller import Signal


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.workflow.archiver')


class Archiver(object):
    def __init__(self, client, workflow, instance):
        self._client = client
        self._workflow = workflow
        self._instance = instance

    def _get_instance_tokens(self):
        """Retrieve all workflow instance tokens.

        Returns:
            List of tokens in the workflow instance.
        """
        prefix = Name(workflow=self._workflow, instance=self._instance)
        query = Query(namePrefix=prefix.get_instance_prefix())
        query_request = QueryRequest(queries=[query])
        try:
            query_response = self._client.query(query_request)
        except TokenMasterException:
            LOG.exception('error sending request %s', query_request)
            return None
        if not query_response.tokens:
            return None
        assert len(query_response.tokens) == 1
        return query_response.tokens[0]

    def _archive_tokens(self, tokens):
        """Archive tokens.

        Args:
            tokens: The list of tokens to archive.
        Returns:
            True iff tokens have been successfully archived.
        """
        archive_request = ArchiveRequest(tokens=tokens)
        try:
            self._client.archive(archive_request)
        except TokenMasterException:
            # It's no big deal if the request failed this time - the same or a
            # different worker will try it again some other time.
            LOG.exception('error sending request %s', archive_request)
            return False
        return True

    def archive_if_expired(self, expiration_timestamp):
        """Archive workflow instance tokens if we passed expiration time.

        Args:
            expiration_timestamp: Timestamp that has to be reached before
                archiving tokens.
        Returns:
            True iff the workflow was successfully archived.
        """
        if expiration_timestamp > time.time():
            return False
        workflow_tokens = self._get_instance_tokens()
        if not workflow_tokens:
            return False
        return self._archive_tokens(workflow_tokens)

    def _has_abort_token(self, tokens):
        """Check if a list of tokens contains an abort token.

        Args:
            tokens: The list of tokens to check.
        Returns:
            True iff the list contains an abort token.
        """
        abort_signal = Signal.action_to_string(Signal.ABORT)
        abort_name = Name(workflow=self._workflow,
                          instance=self._instance,
                          signal=abort_signal)
        abort_token_name = abort_name.get_signal_token_name()
        for token in tokens:
            if token.name == abort_token_name:
                return True
        return False

    @staticmethod
    def _is_owned(token):
        """Check if a given token is owned.

        Returns:
            True iff the given token is owned. Due to a possible clock skew
            between machines it is hard to tell with a high confidence if a
            token is owned at a given time.  We prefer to err on the false
            positive side - i.e., if the result is False, we are highly
            confident that the token is not owned, but not the opposite.
        """
        # Clocks can be off by this much on different machines.
        CLOCK_SKEW_THRESHOLD_SEC = 10
        timestamp = time.time()
        if not token.expirationTime:
            return False
        return timestamp - token.expirationTime < CLOCK_SKEW_THRESHOLD_SEC

    @staticmethod
    def _has_owned_tokens(tokens):
        """Check if any token in a given list is owned.

        Args:
            tokens: The tokens whose ownership should be assessed.
        Returns:
            True iff the list contains an owned token.
        """
        for token in tokens:
            if Archiver._is_owned(token):
                return True
        return False

    def archive_if_aborted(self):
        """Archives workflow instance if it has been aborted.

        Returns:
            True iff the workflow has been aborted.
        """
        workflow_tokens = self._get_instance_tokens()
        if (not workflow_tokens or
                not self._has_abort_token(workflow_tokens) or
                Archiver._has_owned_tokens(workflow_tokens)):
            return False
        return self._archive_tokens(workflow_tokens)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for our workflow config."""
import unittest

from pinball_ext.examples.workflows import WORKFLOWS


__author__ = 'Pawel Garbacki, Mao Ye, Jooseong Kim'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class WorkflowConfigTestCase(unittest.TestCase):
    def test_pinball_config_parses(self):
        from pinball_ext.workflow.parser import PyWorkflowParser
        params =\
            {'workflows_config': 'pinball_ext.examples.workflows.WORKFLOWS'}
        py_workflow_parser = PyWorkflowParser(params)
        py_workflow_parser.parse_workflows()

    def test_duplicate_job_definition(self):
        jobs = {}
        for workflow_name, workflow in WORKFLOWS.items():
            for job_name, job_config in workflow.jobs.items():
                if not getattr(job_config.template, '_REUSABLE', False):
                    if jobs.get(job_name):
                        raise Exception(
                            "%s is declared in both workflow %s and %s." %
                            (job_name, workflow_name, jobs[job_name]))
                    else:
                        jobs[job_name] = workflow_name



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for transactions."""
import copy
import pickle
import pytrie
import sys
import unittest

from pinball.master.blessed_version import BlessedVersion
from pinball.master.master_handler import MasterHandler
from pinball.master.thrift_lib.ttypes import ArchiveRequest
from pinball.master.thrift_lib.ttypes import GroupRequest
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import Query
from pinball.master.thrift_lib.ttypes import QueryAndOwnRequest
from pinball.master.thrift_lib.ttypes import QueryRequest
from pinball.master.thrift_lib.ttypes import Token
from pinball.master.thrift_lib.ttypes import TokenMasterException
from pinball.master.transaction import ArchiveTransaction
from pinball.master.transaction import GroupTransaction
from pinball.master.transaction import ModifyTransaction
from pinball.master.transaction import QueryAndOwnTransaction
from pinball.master.transaction import QueryTransaction
from tests.pinball.persistence.ephemeral_store import EphemeralStore


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class TransactionTestCase(unittest.TestCase):
    def setUp(self):
        """Set up self._trie with 111 tokens, one of them a blessed version."""
        self._trie = pytrie.StringTrie()
        self._store = EphemeralStore()
        blessed_version = BlessedVersion(MasterHandler._BLESSED_VERSION,
                                         MasterHandler._MASTER_OWNER)
        for i in range(0, 10):
            some_token = Token(blessed_version.advance_version(),
                               '/some_dir/some_token_%d' % i,
                               priority=i,
                               data='some_data_%d' % i)
            self._trie[some_token.name] = some_token
            self._store.commit_tokens(updates=[some_token])
            for j in range(0, 10):
                some_other_token = Token(
                    blessed_version.advance_version(),
                    '/some_dir/some_token_%d/some_other_token_%d' % (i, j),
                    priority=j,
                    data='some_data_%d_%d' % (i, j))
                self._trie[some_other_token.name] = some_other_token
                self._store.commit_tokens(updates=[some_other_token])
        blessed_version.advance_version()
        self._trie[MasterHandler._BLESSED_VERSION] = blessed_version
        self._store.commit_tokens(updates=[blessed_version])
        self._check_version_uniqueness()

    def _check_version_uniqueness(self):
        """Check self._trie.values() have distinct version values."""
        versions = set()
        for token in self._trie.values():
            versions.add(token.version)
        self.assertEqual(len(self._trie), len(versions))

    def _get_blessed_version(self):
        return self._trie[MasterHandler._BLESSED_VERSION]

    # Archive tests.
    def test_archive_empty(self):
        request = ArchiveRequest()
        transaction = ArchiveTransaction()
        # Make sure that prepare and commit do not throw an exception.
        transaction.prepare(request)
        transaction.commit(self._trie,
                           self._get_blessed_version(),
                           self._store)

    def test_archive(self):
        request = ArchiveRequest(tokens=[])
        n_tokens_before = len(self._trie)
        some_token = copy.copy(self._trie['/some_dir/some_token_0'])
        request.tokens.append(some_token)
        some_other_token = copy.copy(
            self._trie['/some_dir/some_token_0/some_other_token_0'])
        request.tokens.append(some_other_token)
        transaction = ArchiveTransaction()
        transaction.prepare(request)
        transaction.commit(self._trie,
                           self._get_blessed_version(),
                           self._store)
        n_tokens_after = len(self._trie)
        # We deleted two things from self._trie.
        self.assertEqual(n_tokens_before - 2, n_tokens_after)
        n_active_tokens = len(self._store.read_active_tokens())
        self.assertEqual(n_tokens_after, n_active_tokens)
        n_all_tokens = len(self._store.read_tokens())
        self.assertEqual(n_tokens_before, n_all_tokens)

    # Group tests.
    def test_group_empty(self):
        request = GroupRequest()
        transaction = GroupTransaction()
        # Make sure that prepare and commit do not throw an exception.
        transaction.prepare(request)
        transaction.commit(self._trie,
                           self._get_blessed_version(),
                           self._store)

    def test_group(self):
        request = GroupRequest()
        request.namePrefix = '/some_dir/'
        request.groupSuffix = '/'
        transaction = GroupTransaction()
        transaction.prepare(request)
        response = transaction.commit(self._trie,
                                      self._get_blessed_version(),
                                      self._store)

        expected_groups = set()
        for i in range(0, 10):
            expected_groups.add('/some_dir/some_token_%d' % i)
        groups = set()
        for group, count in response.counts.iteritems():
            groups.add(group)
            self.assertEqual(11, count)
        self.assertEqual(expected_groups, groups)

    # Modify tests.
    def test_modity_empty(self):
        request = ModifyRequest()
        transaction = ModifyTransaction()
        # Make sure that prepare and commit do not throw an exception.
        transaction.prepare(request)
        transaction.commit(self._trie,
                           self._get_blessed_version(),
                           self._store)

    def test_modify_deletes(self):
        request = ModifyRequest(deletes=[])
        n_tokens_before = len(self._trie)
        some_token = copy.copy(self._trie['/some_dir/some_token_0'])
        request.deletes.append(some_token)
        some_other_token = copy.copy(
            self._trie['/some_dir/some_token_0/some_other_token_0'])
        request.deletes.append(some_other_token)
        transaction = ModifyTransaction()
        transaction.prepare(request)
        transaction.commit(self._trie,
                           self._get_blessed_version(),
                           self._store)
        n_tokens_after = len(self._trie)
        # We deleted two things from self._trie.
        self.assertEqual(n_tokens_before - 2, n_tokens_after)
        self._check_version_uniqueness()

    def test_modify_updates(self):
        request = ModifyRequest(updates=[])
        n_tokens_before = len(self._trie)
        token = copy.copy(self._trie['/some_dir/some_token_0'])
        token.data = 'some other data'
        request.updates.append(token)
        new_token = Token(name='/some_other_dir/some_token', data='some data')
        request.updates.append(new_token)
        transaction = ModifyTransaction()
        transaction.prepare(request)
        response = transaction.commit(self._trie,
                                      self._get_blessed_version(),
                                      self._store)

        self.assertEqual(2, len(response.updates))
        self.assertNotEqual(token.version, response.updates[0].version)
        self.assertEqual(token.name, response.updates[0].name)
        self.assertEqual(token.data, response.updates[0].data)
        self.assertLess(0, response.updates[1].version)
        self.assertEqual(new_token.name, response.updates[1].name)
        self.assertEqual(new_token.data, response.updates[1].data)
        n_tokens_after = len(self._trie)
        self.assertEqual(n_tokens_before + 1, n_tokens_after)
        self._check_version_uniqueness()

    def test_modify_no_name_change(self):
        """Do not allow changing token names."""
        request = ModifyRequest(updates=[])
        # pickle gets maximum recursion depth exceeded when traversing
        # the trie, probably a bug in pickle. Setting the recursion limit
        # to a high number fixes it.
        sys.setrecursionlimit(10000)
        trie_before = pickle.dumps(self._trie)
        token = copy.copy(self._trie['/some_dir/some_token_0'])
        token.name = '/some_other_dir/some_token_0'
        request.updates.append(token)
        transaction = ModifyTransaction()
        transaction.prepare(request)
        self.assertRaises(TokenMasterException, transaction.commit,
                          self._trie, self._get_blessed_version(), self._store)
        trie_after = pickle.dumps(self._trie)
        self.assertEqual(trie_before, trie_after)

    def test_modify_deletes_and_updates(self):
        """Updates and deletes in a single request."""
        request = ModifyRequest(updates=[], deletes=[])
        n_tokens_before = len(self._trie)
        delete_token = copy.copy(self._trie['/some_dir/some_token_0'])
        request.deletes.append(delete_token)
        update_token = copy.copy(self._trie['/some_dir/some_token_1'])
        update_token.data = 'some other data'
        request.updates.append(update_token)
        transaction = ModifyTransaction()
        transaction.prepare(request)
        response = transaction.commit(self._trie,
                                      self._get_blessed_version(),
                                      self._store)

        self.assertEqual(1, len(response.updates))
        n_tokens_after = len(self._trie)
        self.assertEqual(n_tokens_before - 1, n_tokens_after)
        self._check_version_uniqueness()

    # Query tests.
    def test_query_empty(self):
        request = QueryRequest()
        transaction = QueryTransaction()
        # Make sure that prepare and commit do not throw an exception.
        transaction.prepare(request)
        transaction.commit(self._trie,
                           self._get_blessed_version(),
                           self._store)

    def test_query(self):
        some_query = Query()
        some_query.namePrefix = '/some_dir'
        some_query.maxTokens = 10
        some_other_query = Query()
        some_other_query.namePrefix = '/some_dir/some_token_0'
        some_other_query.maxTokens = 100
        request = QueryRequest()
        request.queries = [some_query, some_other_query]
        transaction = QueryTransaction()
        transaction.prepare(request)
        response = transaction.commit(self._trie,
                                      self._get_blessed_version(),
                                      self._store)
        self.assertEqual(2, len(response.tokens))
        self.assertEqual(10, len(response.tokens[0]))
        for token in response.tokens[0]:
            self.assertTrue(token.name.startswith('/some_dir'))
            self.assertEqual(9, token.priority)
        self.assertEqual(11, len(response.tokens[1]))
        for token in response.tokens[1]:
            self.assertTrue(token.name.startswith('/some_dir/some_token_0'))

    # Query and own tests.
    def test_query_and_own_empty(self):
        request = QueryAndOwnRequest()
        transaction = QueryAndOwnTransaction()
        # Make sure that prepare and commit do not throw an exception.
        transaction.prepare(request)
        transaction.commit(self._trie,
                           self._get_blessed_version(),
                           self._store)

    def test_query_and_own(self):
        some_token = self._trie['/some_dir/some_token_0']
        some_token.owner = 'some_owner'
        some_token.expirationTime = 10  # in the past
        some_token = self._trie['/some_dir/some_token_1']
        some_token.owner = 'some_owner'
        some_token.expirationTime = sys.maxint  # in the future
        some_query = Query()
        some_query.namePrefix = ''
        some_query.maxTokens = 200
        request = QueryAndOwnRequest()
        request.owner = 'some_other_owner'
        request.expirationTime = sys.maxint
        request.query = some_query
        transaction = QueryAndOwnTransaction()
        transaction.prepare(request)
        response = transaction.commit(self._trie,
                                      self._get_blessed_version(),
                                      self._store)

        # Should have owned all tokens but two: the blessed version and the one
        # token that is already owned.
        self.assertEqual(len(self._trie) - 2, len(response.tokens))
        for token in response.tokens:
            self.assertEquals('some_other_owner', token.owner)
            self.assertEquals(sys.maxint, token.expirationTime)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data models describing tokens.
"""
from django.db import models

from pinball.config.utils import master_name
from pinball.master.thrift_lib.ttypes import Token


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class TokenModel(models.Model):
    """Token model stored as a row in database table."""
    version = models.BigIntegerField()
    # TODO(pawel): 255 chars may not be sufficient to represent longer names
    # but this is a limit for unique char fields:
    # https://docs.djangoproject.com/en/dev/ref/databases/#character-fields
    # Consider alternatives.
    name = models.CharField(max_length=255, primary_key=True)
    owner = models.CharField(max_length=255, blank=True, null=True)
    expirationTime = models.BigIntegerField(blank=True, null=True)
    priority = models.FloatField(blank=True, null=True)
    data = models.TextField(blank=True, null=True)

    def __unicode__(self):
        # Modifying the return value will break the data retrieval logic
        # in the implementation of the DbStore.
        return self.name

    def to_token(self):
        return Token(version=self.version,
                     name=self.name,
                     owner=self.owner,
                     expirationTime=self.expirationTime,
                     priority=self.priority,
                     data=self.data)

    class Meta:
        abstract = True


class ActiveTokenModel(TokenModel):
    """Active token represents a token stored in a pinball master."""
    @staticmethod
    def from_token(token):
        token_model = ActiveTokenModel(version=token.version,
                                       name=token.name,
                                       owner=token.owner,
                                       expirationTime=token.expirationTime,
                                       priority=token.priority,
                                       data=token.data)
        return token_model

    # TODO(mao): refactor code to remove the duplicated table name creation.
    @staticmethod
    def get_table_name():
        return 'active_tokens_%s' % master_name()

    class Meta(TokenModel.Meta):
        # Table name is suffixed with master name.  This way multiple masters
        # can store their state in the same database.
        db_table = 'active_tokens_%s' % master_name()


class ArchivedTokenModel(TokenModel):
    """Archived token is not stored in pinball master any more."""
    @staticmethod
    def from_token(token):
        return ArchivedTokenModel(version=token.version,
                                  name=token.name,
                                  owner=token.owner,
                                  expirationTime=token.expirationTime,
                                  priority=token.priority,
                                  data=token.data)

    @staticmethod
    def get_table_name():
        return 'archived_tokens_%s' % master_name()

    class Meta(TokenModel.Meta):
        db_table = 'archived_tokens_%s' % master_name()


class CachedDataModel(models.Model):
    """Cache data model stored as a row in database table."""
    name = models.CharField(max_length=255, primary_key=True)
    data = models.TextField(blank=True, null=True)

    def __unicode__(self):
        return self.name

    @staticmethod
    def from_data(name, data):
        return CachedDataModel(name=name, data=data)

    @staticmethod
    def get_table_name():
        return 'cached_data_%s' % master_name()

    class Meta:
        db_table = 'cached_data_%s' % master_name()



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import mock
from pinball_ext.executor.common import Platform
from pinball_ext.job import hadoop_jobs


__author__ = 'Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'

_HADOOP_TEST_JOB_CLASS_NAME = '_NON_EXISTS_CLASS_NAME_'

_PARAMS = {
    'start_date': '2015-01-01',
    'end_date': '2015-01-07',
}

_SETTINGS = {
    'executor': Platform.QUBOLE,
    'executor_config': {
        'USER_APPJAR_PATH': 's3n:/my_bucket/my_dir/my_jar.jar',
    }
}

_JOB_STDOUT = 'stdout'
_JOB_STDERR = 'stderr'
_JOB_IDS = ['9527']

_JOBCONF_ARGS = {
   'job_conf_arg1': 'value1',
   'job_conf_arg2': 'value2',
}

_EXTRA_JARS = ['s3n://my_bucket/my_dir/myjar1.jar',
               's3n://my_bucket/my_dir/myjar2.jar']


class HadoopTestJob(hadoop_jobs.HadoopJob):
    def _setup(self):
        self.jobconf_args = _JOBCONF_ARGS
        self.extra_jars = _EXTRA_JARS
        self.extra_arguments = ['-arg value']

    def _get_class_name(self):
        return _HADOOP_TEST_JOB_CLASS_NAME


def _mock_qubole_run_hadoop_job(class_name,
                                jobconf_args,
                                extra_args,
                                extra_jars):
    return _JOB_STDOUT, _JOB_STDERR, _JOB_IDS


class HadoopJobTestCase(unittest.TestCase):
    @mock.patch('pinball_ext.executor.qubole_executor.QuboleExecutor.__init__',
                mock.Mock(return_value=None))
    @mock.patch('pinball_ext.executor.qubole_executor.QuboleExecutor.run_hadoop_job',
                side_effect=_mock_qubole_run_hadoop_job)
    def test_hadoop_job(self, f1):
        hjob = HadoopTestJob(params=_PARAMS, settings=_SETTINGS)
        hjob.runjob()
        self.assertEquals(hjob._job_output, _JOB_STDOUT)
        self.assertEquals(hjob._job_stderr, _JOB_STDERR)
        self.assertEquals(hjob._job_ids, _JOB_IDS)
        f1.assert_called_once_with(
            _HADOOP_TEST_JOB_CLASS_NAME,
            jobconf_args=_JOBCONF_ARGS,
            extra_args=['-start_date 2015-01-01', '-end_date 2015-01-07', '-arg value'],
            extra_jars=_EXTRA_JARS)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Singleton classes needed for authentication."""

import binascii
import json
from re import compile

from Crypto.Cipher import AES, ARC2
from Crypto.Hash import HMAC, SHA256
from Crypto.Random import random
from django.http import HttpResponseRedirect
from oauth2client.client import OAuth2WebServerFlow

from pinball.config.pinball_config import PinballConfig


__author__ = 'Tongbo Huang, Devin Lundberg'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = ['Tongbo Huang', 'Devin Lundberg']
__license__ = 'Apache'
__version__ = '2.0'


class SingletonType(type):
    """This is a singleton metaclass.

    All classes created by this metaclass will only be initialized once.
    """
    def __call__(cls, *args, **kwargs):
        try:
            return cls.__instance
        except AttributeError:
            cls.__instance = super(SingletonType, cls).__call__(*args, **kwargs)
            return cls.__instance


class OAuth2Flow(object):
    """This class is used to create a google OAuth2WebServerFlow object with
    Pinball client credentials.

    You can created new ones from Google Cloud Console.
    """
    __metaclass__ = SingletonType
    _scope = 'https://www.googleapis.com/auth/userinfo.email '\
             'https://www.googleapis.com/auth/admin.directory.group.readonly'
    _redirect_uri = ''
    if not PinballConfig.UI_HOST:
        _redirect_uri = 'http://localhost:8080/oauth2callback/'
    else:
        _redirect_uri = 'https://%s/oauth2callback/' % PinballConfig.UI_HOST
    _flow = {}
    for domain in PinballConfig.AUTHENTICATION_DOMAINS:
        _flow[domain] = OAuth2WebServerFlow(client_id=PinballConfig.GOOGLE_CLIENT_ID,
                                            client_secret=PinballConfig.GOOGLE_CLIENT_SECRET,
                                            scope=_scope,
                                            redirect_uri=_redirect_uri,
                                            hd=domain)

    def get_flow(self, domain):
        """Getter for the google OAuth2WebServerFlow object with
        Pinball client credentials.

        Returns:
            The google OAuth2WebServerFlow object with Pinball client
            credentials.
        """
        return self._flow[domain]

    def domain_authenticated(self, domain):
        """Check if the given domain is authorized with the application.

        Returns:
            The domain to be checked.
        """
        return domain in self._flow.keys()

    def get_domains(self):
        """Get all authenticated domains.

        Returns:
            Authenticated domains
        """
        return PinballConfig.AUTHENTICATION_DOMAINS


class CryptoException(Exception):
    pass


class Crypter(object):
    """Implements modern authenticated encryption for strings

    The current version (1) uses HMAC-SHA256 and AES256-CBC using
    Encrypt-then-MAC. Should be secure >>>2030 according to NIST standards.
    http://www.keylength.com/ for a summary of algorithms and expected
    security. The message is padded with null characters. This means that
    strings with null characters should not use this method unless a
    different padding scheme is added

    AES in GCM mode is faster, but the pycrypto implementation is immature.
    This may be a better choice for future versions.

    There is currently legacy support for decryption under the old key
    using ARC2
    TODO(devinlundberg): remove legacy ARC2 decryption support
    """
    __metaclass__ = SingletonType
    # Remove legacy crypter in future.
    _legacy_crypter = ARC2.new(PinballConfig.SECRET_KEY, ARC2.MODE_ECB)
    _aes_block_size = 16
    _padding_char = '\x00'

    def _serialize(self, ciphertext, mac, **params):
        """Creates a serialized crypto object with current version and key"""
        encoded_params = {k: v.encode('base64') for k, v in params.items()}
        return json.dumps({
            'version': PinballConfig.CRYPTO_VERSION,
            'ciphertext': ciphertext.encode('base64'),
            'auth': mac.encode('base64'),
            'params': encoded_params
        }).encode('base64')

    def _deserialize(self, encoded_ciphertext):
        """Gets version, ciphertext, auth, and params from serialized object"""
        try:
            ciphertext_json = encoded_ciphertext.decode('base64')
        except binascii.Error:
            raise CryptoException('Invalid Base64')
        try:
            ciphertext_obj = json.loads(ciphertext_json)
        except ValueError:
            raise CryptoException('Invalid JSON format')
        if any(key not in ciphertext_obj
               for key in ('version', 'ciphertext', 'auth', 'params')):
            raise CryptoException('Invalid JSON parameters')
        version = ciphertext_obj['version']
        try:
            ciphertext = ciphertext_obj['ciphertext'].decode('base64')
            auth = ciphertext_obj['auth'].decode('base64')
            params = {k: v.decode('base64')
                      for k, v in ciphertext_obj['params'].items()}
        except binascii.Error:
            raise CryptoException('Invalid Base64')
        except AttributeError:
            raise CryptoException('Unsupported types')
        return version, ciphertext, auth, params

    def _cbc_hmac_sha256_decrypt(self, ciphertext, auth, iv):
        """Authenticated decrypt using AES-CBC and HMAC SHA256
        Encrypt-then-MAC"""
        hmac = HMAC.new(PinballConfig.HMAC_KEY, digestmod=SHA256)
        hmac.update(ciphertext)
        hmac.update(iv)
        if hmac.hexdigest() != auth:
            raise CryptoException('Decryption Failed')
        aes = AES.new(PinballConfig.AES_CBC_KEY, AES.MODE_CBC, iv)
        return aes.decrypt(ciphertext).rstrip(self._padding_char)

    def encrypt(self, message):
        """Encrypts string of any length using the current crypto version

        Args:
            message: The string that needs to be encrypted.

        Returns:
            A serialized authenticated encrypted object
        """
        iv = ''.join(chr(random.randint(0, 255))
                     for _ in range(self._aes_block_size))
        aes = AES.new(PinballConfig.AES_CBC_KEY, AES.MODE_CBC, iv)
        hmac = HMAC.new(PinballConfig.HMAC_KEY, digestmod=SHA256)
        padded_length = (len(message) + self._aes_block_size -
                         len(message) % (self._aes_block_size))
        padded_message = message.ljust(padded_length, self._padding_char)
        ciphertext = aes.encrypt(padded_message)
        hmac.update(ciphertext)
        hmac.update(iv)
        return self._serialize(ciphertext, hmac.hexdigest(), iv=iv)

    def decrypt(self, encoded_ciphertext):
        """Deserializes and decrypts a string with the current or legacy
        algorithms

        Args:
            encoded_ciphertext: The string that needs to be decrypted.

        Returns:
            The decrypted message.

        Throws:
            CryptoException: on failed decryption.
        """
        try:
            version, ciphertext, auth, params = self._deserialize(encoded_ciphertext)
        except CryptoException:
            # This should raise an exception when support for ARC2 ends.
            return self._legacy_crypter.decrypt(encoded_ciphertext).rstrip('0')
        if version == 1:
            if 'iv' not in params:
                raise CryptoException('Missing IV')
            return self._cbc_hmac_sha256_decrypt(ciphertext, auth, params['iv'])
        else:
            raise CryptoException('Unsupported Crypto Version')


class RequireLogin(object):
    """This middleware requires a user to be authenticated to view any pages.

    Exemptions to this requirement can optionally be specified
    in settings via a list of regular expressions in EXEMPT_URLS (which
    you can copy from your urls.py).

    Requires authentication middleware and template context processors to be
    loaded. You'll get an error if they aren't.
    """
    def process_request(self, request):
        """The Login Required middleware requires authentication middleware to
        be installed. Edit your MIDDLEWARE_CLASSES setting to insert
        'django.contrib.auth.middlware.AuthenticationMiddleware'.
        If that doesn't work, ensure your TEMPLATE_CONTEXT_PROCESSORS setting
        includes 'django.core.context_processors.auth'.

        Args:
            request: The request sent towards any application url.

        Returns:
            Redirect user to signin page if the user has not logged in or their
            client side credential cookie is malform.
        """

        domains = OAuth2Flow().get_domains()
        EXEMPT_URLS = [compile('/signin/'),
                       compile('/oauth2callback/'),
                       compile('/logout/'),
                       compile('/static/.*')]
        path = request.path_info
        if any(m.match(path) for m in EXEMPT_URLS):
            pass
        elif 'login' in request.COOKIES and 'user_id' in request.COOKIES \
                and 'domain_url' in request.COOKIES:
            crypter = Crypter()
            try:
                user_id = crypter.decrypt(request.COOKIES['login'])
                domain = crypter.decrypt(request.COOKIES['domain_url'])
            except CryptoException:
                return HttpResponseRedirect('/signin/')
            if user_id == request.COOKIES['user_id'] and user_id != '' \
                    and domain in domains:
                pass
            else:
                return HttpResponseRedirect('/signin/')
        else:
            return HttpResponseRedirect('/signin/')



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An executor for running Hadoop/Hive jobs in Qubole cluster.

For more information about Qubole and its API, please check:
    - https://api.qubole.com
"""

import getpass
import logging
import os
import re
import subprocess
import sys
import StringIO
import time

from qds_sdk.commands import Command
from qds_sdk.commands import HiveCommand
from qds_sdk.commands import ShellCommand
from qds_sdk.qubole import Qubole
from pinball_ext.common import output_filter
from pinball_ext.common import s3_utils
from pinball_ext.common import utils
from pinball_ext.common.decorators import retry
from pinball_ext.executor.cluster_executor import ClusterExecutor
from pinball_ext.executor.common import Platform


__author__ = 'Zach Drach, Changshu Liu, Jie Li'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class QuboleExecutor(ClusterExecutor):
    """Implements the executor on the Qubole platform.

    The implementation is based on the Qubole qds_sdk Python version. For
    detailed information, please check: https://github.com/qubole/qds-sdk-py.

    The following configuration paths are supposed to be absolute paths on s3:
        - self.config.USER_LIBJAR_DIRS
        - self.config.USER_APPJAR_PATH
        - self.config.USER_ARCHIVE_PATH

    This executor will generate a shell command line, which downloads all
    resource files to a node inside Qubole cluster and then submits the actual
    Hadoop job to the cluster from that node.

    TODO(csliu):
     - make executor work when jars are stored on HDFS.
     - make Hadoop streaming work for this executor.
    """

    class Config(ClusterExecutor.Config):
        # Qubole API token required to use Qubole's API, check
        # https://api.qubole.com/users/edit for detail.
        if getpass.getuser() == 'prod':
            # Should be configured in executor ctor later.
            API_TOKEN = ''
        else:
            # Make it flexible if developer runs it manually.
            API_TOKEN = os.environ.get('QUBOLE_API_TOKEN')

        API_URL = 'https://api.qubole.com/api/'
        API_VERSION = 'v1.2'
        POLL_INTERVAL_SEC = 20

        SCHEDULER_PARAM = 'mapred.fairscheduler.pool'
        PLATFORM = Platform.QUBOLE

        # These jars are already provided in Qubole's Hadoop distribution, so we need
        # to remove these jars from the Hadoop classpath, and also not ship these jars
        # over to the task nodes.
        QUBOLE_JARS_BLACKLIST = [
            'slf4j-api-1.7.2.jar',
            'slf4j-jdk14-1.6.4.jar',
            'slf4j-log4j12-1.7.2.jar',
            'log4j-1.2.15.jar',
            'log4j-over-slf4j-1.7.2.jar',
            'hadoop-core-1.0.3.jar',
            'hadoop-lzo-0.4.16.jar',
            'hadoop-test-1.0.3.jar',
            'hadoop-tools-1.0.3.jar',
            'jcl-over-slf4j-1.7.2.jar',
        ]

        # Number of retries for pulling status info from Quoble.
        NUM_RETRIES = 6

        # Initial delay for the retry of pulling status info from Quoble.
        INITIAL_DELAY = 60

    def __init__(self, executor_config=None):
        qubole_blacklist_jars = None
        for key, value in executor_config.items():
            if key == 'QUBOLE_JARS_BLACKLIST':
                qubole_blacklist_jars = value.split(',')
        if qubole_blacklist_jars:
            executor_config['QUBOLE_JARS_BLACKLIST'] = qubole_blacklist_jars

        super(QuboleExecutor, self).__init__(executor_config=executor_config)

    def run_hive_query(self, query_str, upload_archive=False):
        # tmp is a keyword in Qubole
        regex = r'([\s\(\),;]|^)tmp([\s\(\),;]|$)'
        query_str = re.sub(regex, r'\g<1>`tmp`\g<2>', query_str)

        full_query_string = self._generate_hive_query_header(
            upload_archive=upload_archive)
        full_query_string += self._get_scheduler_hive_setting()
        full_query_string += query_str

        self.log.info('Running query %s' % full_query_string)

        kwargs = dict(query=full_query_string)
        hc, output, stderr, job_ids = self._run_qubole_command_with_stderr(
            HiveCommand, self._hive_query_log_line_processor, kwargs)

        return output, stderr, job_ids

    def kill_command(self, qubole_jid):
        """Kills a qubole job with the given job_id."""
        self._configure_qubole()
        qubole_jid = int(qubole_jid)
        Command.cancel_id(qubole_jid)

    def get_job_result(self, qubole_jid):
        """Finds and retrieves results for existing Qubole job.

        Args:
            id: qubole job id.

        Returns:
            Job stdout output.
        """
        self._configure_qubole()
        qubole_jid = str(qubole_jid)
        return self._get_qubole_command_output(HiveCommand.find(qubole_jid))

    def run_hadoop_job(self,
                       class_name,
                       jobconf_args=None,
                       extra_args=None,
                       extra_jars=None):
        """Run a Hadoop job in Qubole cluster.

        We assume extra_jars are stored on s3 and the path looks like:
            s3://pinball/%{USER}/some_jar_dir/

        We fail the entire command if pulling the JARs down from s3 fails,
        so we use "&&" to connect shell commands.
        """
        jobconf_args = jobconf_args if jobconf_args else {}
        extra_args = extra_args if extra_args else []
        extra_jars = extra_jars if extra_jars else []

        # The place where all jars are stored in s3.
        s3_jar_dirs = self.config.USER_LIBJAR_DIRS + extra_jars
        # The place where all jars will be copied to locally.
        local_jar_dir = '/tmp/hadoop_users/%s/%s' % \
                        (self.config.USER, utils.get_random_string())
        download_jar_cmds = ['hadoop fs -get %s %s' % (s3_dir, local_jar_dir)
                             for s3_dir in s3_jar_dirs]
        download_jar_cmd = ' && '.join(download_jar_cmds)
        appjar_name = s3_utils.extract_file_name_from_s3_path(
            self.config.USER_APPJAR_PATH)
        download_jar_cmd += ' && hadoop fs -get %s %s/%s' % (
            self.config.USER_APPJAR_PATH,
            local_jar_dir,
            appjar_name
        )

        # Set default JobConf args.
        jobconf_args = {} if jobconf_args is None else jobconf_args.copy()
        if self.config.SCHEDULER_QUEUE:
            jobconf_args[self.config.SCHEDULER_PARAM] = \
                self.config.SCHEDULER_QUEUE
        jobconf_args['mapred.job.name'] = self.job_name

        # Create arguments.
        arguments = ' '.join('-D%s=%s' % (k, v) for k, v in jobconf_args.iteritems())
        arguments += ' '
        arguments += ' '.join(extra_args)

        libjars = self._get_libjars_local_paths(s3_jar_dirs, local_jar_dir)
        hadoop_classpath = '%s/*' % local_jar_dir

        cmd = 'mkdir -p %(local_jar_dir)s && %(download_jar_cmd)s'

        files_to_be_deleted = []
        for qubole_jar in self.config.QUBOLE_JARS_BLACKLIST:
            files_to_be_deleted.append('%s/%s' % (local_jar_dir, qubole_jar))
        if files_to_be_deleted:
            cmd += ' && rm -f %s' % (' && rm -f '.join(files_to_be_deleted))

        # Generate command.
        var_dict = {
            'class_name': class_name,
            'arguments': arguments,
            'appjar_name': appjar_name,
            'download_jar_cmd': download_jar_cmd,
            'local_jar_dir': local_jar_dir,
            'hadoop_classpath': hadoop_classpath,
            'libjars': libjars,
        }
        cmd += (' && export HADOOP_CLASSPATH=%(hadoop_classpath)s'
                ' && hadoop jar %(local_jar_dir)s/%(appjar_name)s'
                ' %(class_name)s'
                ' -libjars %(libjars)s'
                ' %(arguments)s')
        cmd += ';\nEXIT_CODE=$?; \nrm -rf %(local_jar_dir)s; \nexit $EXIT_CODE;'
        cmd = cmd % var_dict

        # Log command messages.
        self.log.info('Full command: %s' % cmd)

        # Run command.
        hc, output, stderr, job_ids = self.run_shell_command(cmd)
        return output, stderr, job_ids

    def run_shell_command(self, cmd):
        kwargs = {'inline': cmd}
        hc, output, stderr, job_ids = self._run_qubole_command_with_stderr(
            ShellCommand,
            self._shell_command_log_line_processor,
            kwargs)
        return hc, output, stderr, job_ids

    def _retry_wrapper(self, fn, tries=4):
        return retry(Exception, tries=tries, logger=self.log)(fn)()

    def _get_libjars_local_paths(self, s3_jar_dirs, local_jar_dir):
        """Returns a list of local jar paths downloaded from s3.

        Args:
            s3_jar_dirs: S3 path from which we pull down JARs from.
            local_jar_dir: local path on every machine in the Qubole cluster
                to which jars are pulled down

        Returns:
            List of local file paths as a string, with each file name delimited
            comma (,), if the supplied s3_jar_dirs is valid. Otherwise, returns
            empty string.
        """
        file_paths = []
        for s3_jar_dir in s3_jar_dirs:
            file_paths += s3_utils.list_s3_directory(s3_jar_dir)

        jar_names = [
            s3_utils.extract_file_name_from_s3_path(file_path)
            for file_path in file_paths if str(file_path).endswith('jar')]
        filtered_jar_names = [
            jar_name
            for jar_name in jar_names if jar_name not in self.config.QUBOLE_JARS_BLACKLIST]

        # dedup jar lists.
        filtered_jar_names = list(set(filtered_jar_names))

        final_jar_paths = [
            '%s/%s' % (local_jar_dir, jar_name)
            for jar_name in filtered_jar_names]
        return ','.join(final_jar_paths)

    ############################################################################
    # Private helper functions for Qubole functions
    ############################################################################

    def _configure_qubole(self):
        logging.basicConfig(level=logging.INFO)
        logger = logging.getLogger('qds_connection')
        logger.propagate = False
        qdslog = logging.getLogger('qds')
        if not self.config.API_TOKEN:
            raise Exception("You didn't specify your QUBOLE_API_TOKEN in "
                            "your environment before running commands on "
                            "Qubole!\n. It can be found at http://api.qubole"
                            ".com/users/edit")
        Qubole.configure(api_token=self.config.API_TOKEN,
                         api_url=self.config.API_URL,
                         version=self.config.API_VERSION,
                         poll_interval=self.config.POLL_INTERVAL_SEC)
        return qdslog

    def _get_qubole_command_output(self, q_cmd):
        """Return the stdout output from a Qubole command object.

        Args:
            hc: the qubole command object

        Returns:
            Query stdout output.
        """
        raw_result_str = ''
        file_handler = StringIO.StringIO()
        if q_cmd.status == 'done':
            self._retry_wrapper(lambda: q_cmd.get_results(file_handler), tries=10)
            raw_result_str = file_handler.getvalue()
        file_handler.close()

        rows = []
        # Qubole writes very large outputs to S3,
        # with each field in this file delimited by ^A
        raw_result_str = raw_result_str.replace('\x01', '\t')
        lines = raw_result_str.strip().split('\n')

        for line in lines:
            if line == '[Empty]':
                continue
            if line.strip():
                rows.append(line.strip().split('\t'))

        return rows

    def _run_qubole_command_with_stderr(self, cls, log_line_processor, kwargs):
        """Run the Qubole command and print the log to sys.stderr in real time.

        Args:
            cls: the qds_sdk.command.Command subclass
            kwargs: for the constructor of the qubole command class

        Returns:
            The tuple (hc, output, stderr)

            hc: the qubole command object pointer
            output: the command output, same as run_raw_hive_query
            stderr: the command stderr, same as run_raw_hive_query
        """
        self._configure_qubole()

        # TODO(mao): set proper number for the tries param.
        hc = self._retry_wrapper(lambda: cls.create(**kwargs), tries=10)
        sys.stderr.write("PINBALL:kill_id=%s/%s\n" % (self.config.PLATFORM,
                                                      hc.id))
        sys.stderr.flush()

        f = None
        if log_line_processor:
            f = output_filter.OutputFilter(log_line_processor,
                                           output=sys.stderr)
        stderr_list = []
        self._job_ids = []

        retry_exception = self.config.NUM_RETRIES
        retry_delay = self.config.INITIAL_DELAY
        while retry_exception > 0:
            try:
                if cls.is_done(hc.status):
                    break
                self.log.info("Sleeping for %s seconds and polling."
                              % Qubole.poll_interval)
                time.sleep(Qubole.poll_interval)
                hc = cls.find(hc.id)

                # TODO(csliu): polling entire error log file is very inefficient
                stderr = self._retry_wrapper(lambda: hc.get_log())
                stderr = stderr.strip().split('\n')
                for i in range(len(stderr_list), len(stderr)):
                    line = stderr[i] + "\n"
                    stderr_list.append([line])
                    if f:
                        f.process_and_output([line])

                # Get a successful status pulling from quoble, reset the retry
                # exception number
                retry_exception = self.config.NUM_RETRIES
            except Exception as e:
                retry_exception -= 1
                self.log.error("Got error %s when checking Qubole status."
                               " Going to retry %d more times." %
                               (e.message, retry_exception))
                time.sleep(retry_delay)
                retry_delay *= 2

        query_id = str(hc.id)
        self.log.info('Completed Query, Id: %s, Status: %s' %
                      (query_id, hc.status))

        if hc.status == 'error' or hc.status == 'cancelled':
            error_message = "Failed on query: %s" % query_id
            raise subprocess.CalledProcessError(1, error_message)
        elif hc.status == "running":
            error_message = "The job is still running, but got too many " \
                            "qubole exceptions: %s" % query_id
            raise subprocess.CalledProcessError(1, error_message)

        self.log.info("Now receiving the query output.")
        output = self._get_qubole_command_output(hc)
        self.log.info("Received the query output.")
        self.log.info("Output has %d rows. First 10 rows:\n\t%s"
                      % (len(output),
                         '\n\t'.join([str(o) for o in output[:9]])))

        return hc, output, stderr_list, self._job_ids

    ############################################################################
    # Private log line processors.
    ############################################################################

    def _hive_query_log_line_processor(self, line):
        """ A callback function that gets executed for every line of
        stderr coming from the running job. Returns a dict of pinball
        metadata.
        """
        job_regex = r"Starting Job = (?P<job_id>.+?), " \
                    r"Tracking URL = <a href='(?P<job_url>.+?)'"
        m = re.search(job_regex, line)
        if m:
            job_id = m.group('job_id')
            job_url = m.group('job_url')
            if job_id and job_url:
                self._job_ids.append(job_id)
                return {'job_id': job_id, 'job_url': job_url}
        return {}

    def _shell_command_log_line_processor(self, line):
        """ A callback function that gets executed for every line of
        stderr coming from the running job. Returns a dict of pinball
        metadata.
        """
        job_regex = r"Submitted job: (?P<job_id>.+)"
        m = re.search(job_regex, line)
        if m:
            job_id = m.group('job_id')
            if job_id not in self._job_ids:
                self._job_ids.append(job_id)
            return {'job_id': job_id}

        # Checking "Running job" besides of "Submitted job" makes
        # job_id extraction more reliable even we miss some logs.
        job_regex = r"Running job: (?P<job_id>.+)"
        m = re.search(job_regex, line)
        if m:
            job_id = m.group('job_id')
            if job_id not in self._job_ids:
                self._job_ids.append(job_id)
            return {'job_id': job_id}

        url_regex = r"Tracking URL: <a href='(?P<job_url>.+?)'"
        m = re.search(url_regex, line)
        if m:
            job_url = m.group('job_url')
            return {'job_url': job_url}
        return {}



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for master handler."""
import sys
import unittest

from pinball.master.master_handler import MasterHandler
from pinball.master.thrift_lib.ttypes import ArchiveRequest
from pinball.master.thrift_lib.ttypes import GroupRequest
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import Query
from pinball.master.thrift_lib.ttypes import QueryAndOwnRequest
from pinball.master.thrift_lib.ttypes import QueryRequest
from pinball.master.thrift_lib.ttypes import Token
from tests.pinball.persistence.ephemeral_store import EphemeralStore


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class MasterHandlerTestCase(unittest.TestCase):
    def _insert_token(self, handler):
        request = ModifyRequest()
        token = Token(name='/some_other_dir/some_token', data='some data')
        request.updates = [token]
        response = handler.modify(request)
        self.assertEqual(1, len(response.updates))
        return response.updates[0]

    def test_archive(self):
        handler = MasterHandler(EphemeralStore())
        token = self._insert_token(handler)
        request = ArchiveRequest()
        request.tokens = [token]
        handler.archive(request)
        # The logic handling the request is tested thoroughly in
        # transaction tests.  Here we only make sure that the plumbing is in
        # place.

    def test_group(self):
        request = GroupRequest()
        request.namePrefix = '/'
        handler = MasterHandler(EphemeralStore())
        response = handler.group(request)
        self.assertEqual(1, len(response.counts))
        self.assertEqual(1, response.counts.values()[0])

    def test_modify(self):
        handler = MasterHandler(EphemeralStore())
        self._insert_token(handler)

    def test_query(self):
        query = Query()
        query.namePrefix = ''
        query.maxTokens = 10
        request = QueryRequest()
        request.queries = [query]
        handler = MasterHandler(EphemeralStore())
        response = handler.query(request)
        self.assertEqual(1, len(response.tokens))

    def test_query_and_own(self):
        query = Query()
        query.namePrefix = ''
        query.maxTokens = 10
        request = QueryAndOwnRequest()
        request.owner = 'some_owner'
        request.expirationTime = sys.maxint
        request.query = query
        handler = MasterHandler(EphemeralStore())
        response = handler.query_and_own(request)
        self.assertEqual(0, len(response.tokens))



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for executing commands in a Hadoop cluster.

Commands are executed by SSHing to the Hadoop cluster and performing
the desired command.

The user of scripts which use these functions should either have their SSH key
on the Hadoop cluster's authorized users list or have the hadoop_identity
SSH key pair available to them.
"""

import logging
import os
import re
import sys
import tempfile

from pinball_ext.common import output_filter
from pinball_ext.common import shell_utils


__author__ = 'Dmitry Chechik, Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


LOG = logging.getLogger('common.hadoop_utils')
LOG.setLevel(logging.INFO)

_HADOOP_SSH_OPTS = [
    '-o', 'UserKnownHostsFile=/dev/null',
    '-o', 'StrictHostKeyChecking=no',
    '-o', 'LogLevel=error',
    # Prevent timeout disconnects.
    '-o', 'ServerAliveInterval=60',
]


# Default parameters to connect to a node in a Hadoop cluster.
class HadoopHostConfig(object):
    # User used in SSH.
    USER_NAME = 'hadoop'
    # Host to SSH to.
    HOST_NAME = None
    # Port which we can SSH to.
    SSH_PORT = 22
    # Path for the local ssh identity file.
    SSH_KEY_FILE = '%s/.ssh/hadoop_identity' % \
                   os.environ.get('HOME', '/home/prod')
    # Home path on Hadoop host.
    REMOTE_HADOOP_HOME = '/home/hadoop'

    def __eq__(self, other):
        if type(self) != type(other):
            raise ValueError('comparing apples and carrots (%s)' % type(other))
        return (self.USER_NAME == other.USER_NAME and
                self.HOST_NAME == other.HOST_NAME and
                self.SSH_PORT == other.SSH_PORT and
                self.SSH_KEY_FILE == other.SSH_KEY_FILE and
                self.REMOTE_HADOOP_HOME == other.REMOTE_HADOOP_HOME)


def run_command_in_hadoop(hadoop_node_config,
                          command,
                          cmd_work_dir=None,
                          check_output=False):
    """Run the given command in the Hadoop cluster.

    The caller has to finish reading the output to make sure
    the command completes. This method returns as soon as the command is
    invoked.

    Usage:
        with run_hadoop_command('ls -l') as f:
            result = f.read()

    Args:
        hadoop_node_config - Where to SSH and run this command.
        command - Command to run.
        cmd_work_dir - Path on hadoop cluster to run from.

    Returns:
        A file-like object to the stdout of the command.
    """
    if not cmd_work_dir:
        cmd_work_dir = hadoop_node_config.REMOTE_HADOOP_HOME

    key_file_path = os.path.expanduser(hadoop_node_config.SSH_KEY_FILE)
    if not os.path.exists(key_file_path):
        LOG.warn('SSH key %s does not exist.' % hadoop_node_config.SSH_KEY_FILE)
        key_file_path = None

    return shell_utils.run_ssh_command(
        user_name=hadoop_node_config.USER_NAME,
        host_name=hadoop_node_config.HOST_NAME,
        command=command,
        cmd_work_dir=cmd_work_dir,
        ssh_port=hadoop_node_config.SSH_PORT,
        ssh_options=_HADOOP_SSH_OPTS,
        ssh_key_file=key_file_path,
        check_output=check_output)


def run_and_check_command_in_hadoop(hadoop_node_config,
                                    command,
                                    cmd_work_dir=None,
                                    log_line_processor=None):
    """Run the given command on the Hadoop cluster.

    Args:
        hadoop_node_config - where to SSH to and run this command.
        command - Command to run.
        cmd_work_dir - Path on hadoop cluster to run from.
        log_line_processor - Processor to process the command stdoutput.

    Raises:
        CalledProcessError if it fails.
    """
    if not log_line_processor:
        run_command_in_hadoop(hadoop_node_config, command, cmd_work_dir, check_output=True)
    else:
        command_output = run_command_in_hadoop(hadoop_node_config, command, cmd_work_dir)
        out_filter = output_filter.OutputFilter(log_line_processor,
                                                output=sys.stderr)
        out_filter.read_and_output(command_output)


def hdfs_exists(hadoop_node_config, file_pattern):
    """Return true iff file_pattern exists on HDFS or S3.

    Args:
        hadoop_node_config - where to SSH to and run this command.
        file_pattern - string file name or glob-like pattern.
    """
    cmd = 'hadoop fs -ls %s 2>&1 > /dev/null; echo $?' % file_pattern
    with run_command_in_hadoop(hadoop_node_config, cmd) as f:
        output = f.read()
    return output.strip() == '0'


def hdfs_rmr(file_pattern, safety_check=None, date_partitioned_check=False):
    """Remove recursively a file or file pattern from HDFS or S3.

    Args:
        file_pattern - File to remove.
        safety_check - If present, will check for existence
                       of the string in file_pattern to
                       double-check that we're not deleting
                       an unexpected path.
        date_partitioned_check - If set to true, exception is raised
                       if the directory doesn't end with 'YYYY-MM-DD'
    Raises:
        CalledProcessError on error.
    """
    DATE_PARTITION_REGEX = r'(.*\d{4}-\d{2}-\d{2}/*)'
    if date_partitioned_check:
        if not re.match(DATE_PARTITION_REGEX, file_pattern):
            raise Exception('Attempted to delete non date partitioned dir %s'
                            % file_pattern)

    if safety_check:
        assert(safety_check in file_pattern)
    cmd = 'hadoop fs -rmr %s' % file_pattern
    run_and_check_command_in_hadoop(cmd)


def put_string_to_hadoop(hadoop_node_config, input_string, remote_path):
    """Takes the passed-in string and writes it to a file on a Hadoop node.

    NOTE: we are not putting to HDFS.

    Args:
        input_string - String to write.
        remote_path - Absolute path on Hadoop node's local disk to write to.

    Raises:
        CalledProcessError on error.
    """
    tmp_file = tempfile.NamedTemporaryFile()
    try:
        tmp_file.write(input_string)
        tmp_file.flush()
        shell_utils.scp_file_to_host(
            hadoop_node_config.USER_NAME,
            hadoop_node_config.HOST_NAME,
            tmp_file.name,
            remote_path,
            ssh_options=_HADOOP_SSH_OPTS,
            ssh_port=hadoop_node_config.SSH_PORT,
            ssh_key_file=hadoop_node_config.SSH_KEY_FILE)
    finally:
        tmp_file.close()



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Web server urls."""
from django.conf.urls import include
from django.conf.urls import patterns
from django.conf.urls import url
from django.views.generic import TemplateView
# Uncomment the next two lines to enable the admin:
from django.contrib import admin
from django.http import HttpResponseRedirect

from pinball.config.pinball_config import PinballConfig
from pinball.ui.views import ExecutionView
from pinball.ui.views import ScheduleView
from pinball.ui.views import TokenPathsView
from pinball.ui.views import TokenView
from pinball.ui.views import workflows


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


admin.autodiscover()

urlpatterns = patterns('',
                       url(r'^$',
                           lambda x: HttpResponseRedirect('/workflows/')),
                       url(r'^workflows/$', TemplateView.as_view(
                           template_name='workflows.html'),
                           name='workflows'),
                       url(r'^ajax/workflows/$',
                           workflows,
                           name='ajax/workflows'),
                       url(r'^instances/$',
                           TemplateView.as_view(
                               template_name='instances.html'),
                           name='instances'),
                       url(r'^ajax/instances/$',
                           'pinball.ui.views.instances',
                           name='ajax/instances'),
                       url(r'^jobs/$',
                           TemplateView.as_view(template_name='jobs.html'),
                           name='jobs'),
                       url(r'^ajax/jobs/$', 'pinball.ui.views.jobs',
                           name='ajax/jobs'),
                       url(r'^graph/$', 'pinball.ui.views.graph',
                           name='graph'),
                       url(r'^executions/$',
                           TemplateView.as_view(
                               template_name='executions.html'),
                           name='executions'),
                       url(r'^job_executions/$',
                           TemplateView.as_view(
                               template_name='job_executions.html'),
                           name='job_executions'),
                       url(r'^ajax/executions/$',
                           'pinball.ui.views.executions',
                           name='ajax/executions'),
                       url(r'^execution/$', ExecutionView.as_view(),
                           name='execution'),
                       url(r'^file/$',
                           TemplateView.as_view(template_name='file.html'),
                           name='file'),
                       url(r'^file_content/$',
                           'pinball.ui.views.file_content',
                           name='file_content'),
                       url(r'^schedules/$',
                           TemplateView.as_view(
                               template_name='schedules.html'),
                           name='schedules'),
                       url(r'^ajax/schedules/$',
                           'pinball.ui.views.schedules',
                           name='ajax/schedules'),
                       url(r'^schedule/$', ScheduleView.as_view(),
                           name='schedule'),
                       url(r'^ajax/jobs_from_config/$',
                           'pinball.ui.views.jobs_from_config',
                           name='ajax/jobs_from_config'),
                       url(r'^ajax/command/$',
                           'pinball.ui.views.command',
                           name='ajax/command'),
                       url(r'^token_paths/$', TokenPathsView.as_view(),
                           name='token_paths'),
                       url(r'^ajax/token_paths/$',
                           'pinball.ui.views.token_paths',
                           name='ajax/token_paths'),
                       url(r'^token/$', TokenView.as_view(), name='token'),
                       url(r'^ajax/status/$',
                           'pinball.ui.views.status',
                           name='ajax/status'),
                       # Uncomment the next line to enable the admin:
                       # TODO(pawel): admin interface doesn't work because
                       # of some reason.
                       url(r'^admin/', include(admin.site.urls)),
                       # Authentication.
                       url(r'^signin/$',
                           'pinball.ui.views.signin'),
                       url(r'^oauth2callback/',
                           'pinball.ui.views.auth_return'),
                       url(r'^logout/',
                           'pinball.ui.views.logout'))

if not PinballConfig.DEBUG:  # if DEBUG is True it will be served automatically
    assert len(PinballConfig.STATICFILES_DIRS) == 1
    urlpatterns += patterns('', url(r'^static/(?P<path>.*)$',
                            'django.views.static.serve',
                            {'document_root': PinballConfig.STATICFILES_DIRS[0]}))



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command line tool to start token master."""

import argparse
import sys

from pinball.config.pinball_config import PinballConfig
from pinball.master.factory import Factory


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = ['Pawel Garbacki', 'Mao Ye']
__license__ = 'Apache'
__version__ = '2.0'


def main():
    parser = argparse.ArgumentParser(
        description='Start Pinball master server.')
    parser.add_argument(
        '-c',
        '--config_file',
        dest='config_file',
        required=True,
        help='full path to the pinball setting configure file')
    parser.add_argument(
        '-p',
        '--port',
        dest='port',
        type=int,
        default=PinballConfig.MASTER_PORT,
        help='port to run on')
    options = parser.parse_args(sys.argv[1:])

    PinballConfig.parse(options.config_file)
    master_port = options.port if options.port else PinballConfig.MASTER_PORT
    factory = Factory(master_port=master_port)

    # The reason why these imports are not at the top level is that some of the
    # imported code (db models initializing table names) depends on parameters
    # passed on the command line (master name).  Those imports need to be delayed
    # until after command line parameter parsing.
    from pinball.persistence.store import DbStore
    factory.create_master(DbStore())
    factory.run_master_server()

if __name__ == '__main__':
    main()



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Define the worklfow model.

Workflow is represented as a DAG of jobs. In order to express cross-workflow
job dependencies where a job in a workflow may depend on a job in another
workflow (typically a data dependence), a condition job can be used.
A condition job is a light weight job that can be periodically evaluated to
check whether the output data of an upstream job in another workflow is
available.
"""

import calendar
import pickle

from pinball.config.pinball_config import PinballConfig
from pinball.master.thrift_lib.ttypes import Token
from pinball.parser.config_parser import ConfigParser
from pinball.workflow.event import Event
from pinball.workflow.name import Name
from pinball.workflow.utils import get_unique_workflow_instance
from pinball.workflow.utils import load_path
from pinball_ext.job_templates import JobTemplateBase
from pinball_ext.job_templates import ConditionTemplateBase


__author__ = 'Pawel Garbacki, Mao Ye, Jooseong Kim'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


_NAME_DELIMITER = '.'


def _is_name_qualified(job_name):
    """Check if job name is qualified."""
    return _NAME_DELIMITER in job_name


def _get_qualified_name(workflow_name, job_name):
    """Construct a qualified name from workflow name and job name."""
    return workflow_name + _NAME_DELIMITER + job_name


class PyWorkflowParser(ConfigParser):
    """Workflow config parser implementation.

    Attributes:
        workflows_config: the fully qualified name of the workflows config dict
        workflows: dict from workflow name to WorkflowDef object of the
            workflow

        The object represented by workflows_config is structured as follows:

        {
            workflow_name_0 (string): WorkflowConfig(
                jobs={
                    job_name_0 (string): JobConfig(
                        JobTemplate(...),
                        [parent_job_name_0 (string), ...]
                    ),
                    job_name_1 (string): JobConfig(...),
                    ...
                }
                final_job_config=FinalJobConfig(...),
                schedule=ScheduleConfig(
                    recurrence=recurrence_spec (string),
                    time=start_timestamp (string),
                    start_date=start_date (string),
                    ...),
                notify_emails=email_addresses (string)
            ),
            workflow_name_1 (string): WorkflowConfig(...),

            ...

            workflow_name_2 (string): WorkflowConfig(...),
            workflow_name_3 (string): WorkflowConfig(...),
        },
        ...

        For more documentation on WorkflowConfig, JobConfig and ScheduleConfig,
        refer to workflow/config.py
    """
    def __init__(self, params):
        """Initializes PyWorkflowParser object.

        Args:
            params - a dict contains the following keys:
                'workflows_config': maps to the fully qualified name of the
                    workflows config object described above.
                'job_repo_dir'[optional]: the root dir of the repo where all
                    user jobs are stored.
                'job_import_dirs_config'[optional]: maps to the fully qualified
                    name of the object that stores the list of dirs where user
                    jobs are defined. The dirs are relative path to
                    'job_repo_dir'.
        """
        assert 'workflows_config' in params
        super(PyWorkflowParser, self).__init__(params)

        self.parser_params = params
        self.workflows_config_str = params['workflows_config']
        self.workflows_config = load_path(self.workflows_config_str)
        self.workflows = {}

    def get_schedule_token(self, workflow):
        if not self.workflows:
            self.parse_workflows()
        workflow_def = self.workflows.get(workflow)
        if not workflow_def:
            return None
        return workflow_def.get_schedule_token()

    def get_workflow_tokens(self, workflow):
        if not self.workflows:
            self.parse_workflows()
        workflow_def = self.workflows.get(workflow)
        if not workflow_def:
            return None
        return workflow_def.get_workflow_tokens()

    def get_workflow_names(self):
        if not self.workflows:
            self.parse_workflows()
        return self.workflows.keys()

    def parse_workflows(self):
        """Parse workflow configs converting them to WorkflowDefs."""
        for workflow_name, workflow_config in self.workflows_config.items():
            jobs = {}

            workflow_schedule = self.parse_schedule(workflow_name,
                                                    workflow_config)
            workflow_def = WorkflowDef(workflow_name, workflow_schedule,
                                       workflow_config.notify_emails)
            self.workflows[workflow_name] = workflow_def

            for job_name, job_config in workflow_config.jobs.items():
                # For now, we just pass parser params down to JobDef to generate
                # job template params based on that.
                job_def = JobDef(job_name,
                                 job_config.template,
                                 workflow_def,
                                 self.parser_params)
                workflow_def.add_job(job_def)
                jobs[job_name] = job_def

            # add dependency
            for job_name, job_config in workflow_config.jobs.items():
                child_job = jobs[job_name]
                for parent in job_config.dependency:
                    child_job.add_dep(jobs[parent])

            # add final job
            final_job_def = JobDef('final',
                                   workflow_config.final_job_config.template,
                                   workflow_def)
            leaf_jobs = workflow_def.get_leaf_jobs()

            for leaf_job in leaf_jobs:
                final_job_def.add_dep(leaf_job)

            workflow_def.add_job(final_job_def)

        self._verify_workflows()

    def parse_schedule(self, workflow_name, workflow):
        """Parse schedule config and create workflow schedule."""
        recurrence = workflow.schedule.recurrence.total_seconds()

        overrun_policy = workflow.schedule.overrun_policy
        if overrun_policy is None:
            overrun_policy = OverrunPolicy.SKIP

        max_running_instances = workflow.schedule.max_running_instances
        if max_running_instances is None:
            max_running_instances =\
                PinballConfig.DEFAULT_MAX_WORKFLOW_RUNNING_INSTANCES

        notify_emails = workflow.notify_emails.split(',')

        return WorkflowSchedule(
            next_run_time=int(calendar.timegm(
                workflow.schedule.reference_timestamp.timetuple())),
            recurrence_seconds=recurrence,
            overrun_policy=overrun_policy,
            parser_params=self.parser_params,
            workflow=workflow_name,
            emails=notify_emails,
            max_running_instances=max_running_instances)

    def _verify_workflows(self):
        """Verify all workflows."""
        for workflow in self.workflows.values():
            workflow.verify()


class WorkflowVerificationException(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class WorkflowDef(object):
    """Workflow is composed of jobs."""

    def __init__(self, name, schedule, notify_emails):
        self.name = name
        self.schedule = schedule
        self.notify_emails = notify_emails
        self.jobs = {}

    def __str__(self):
        return ('workflow:%s, schedule:%s, notify_emails:%s, jobs:%s' %
                (self.name, self.schedule, self.notify_emails, str(self.jobs)))

    def __repr__(self):
        return self.__str__()

    def add_job(self, job):
        """Add a job to the workflow."""
        if job.name in self.jobs:
            raise WorkflowVerificationException(
                "job %s already exists in workflow %s" % (job.name, self.name))
        self.jobs[job.name] = job

    def get_leaf_jobs(self):
        """Get all jobs that no one depends on."""
        leaf_jobs = set(self.jobs.values())
        for _, job in self.jobs.items():
            leaf_jobs = leaf_jobs.difference(job.inputs)
        return list(leaf_jobs)

    def _get_top_level_jobs(self):
        all_jobs = self._get_transitive_deps()
        result = []
        for job in all_jobs:
            if not job.inputs or job.workflow.name != self.name:
                result.append(job)
        return result

    def verify(self):
        """Make sure that the workflow is properly structured."""
        self._check_has_single_final_job()
        # make sure that the workflow is connected and acyclic
        self._check_acyclic()

    def _check_has_single_final_job(self):
        """Make sure that the workflow has a single job that no one depends on.
        """
        num_leaf_jobs = len(self.get_leaf_jobs())
        if num_leaf_jobs != 1:
            raise WorkflowVerificationException(
                "expected exactly one final job in workflow %s, but found %i "
                "final jobs" % (self.name, num_leaf_jobs))

    def _get_transitive_deps(self):
        """Find all jobs that the workflow depends on.

        Return all jobs in the workflow and any other job that a job in the
        workflow depends on directly or indirectly.
        """
        def _dfs(job, path, visited):
            visited.add(job)
            if job in path:
                raise WorkflowVerificationException(
                    "job %s is part of a cycle" % job.get_qualified_name())
            path.append(job)
            if job.workflow.name == self.name:
                # Do not traverse deps of jobs that belong to a different
                # workflow.  External dependencies should go one-level deep.
                for dep in job.inputs:
                    _dfs(dep, path, visited)
            path.pop()

        final_jobs = self.get_leaf_jobs()
        assert len(final_jobs) == 1
        visited = set()
        _dfs(final_jobs[0], [], visited)
        return visited

    def _check_acyclic(self):
        """Make sure that job dependencies do not have a cycle."""
        if self.jobs:
            self._get_transitive_deps()

    def get_schedule_token(self):
        """Create a token describing workflow execution schedule."""
        self.schedule.advance_next_run_time()
        timestamp = self.schedule.next_run_time
        token_name = (
            Name(workflow=self.name).get_workflow_schedule_token_name())
        return Token(name=token_name, owner='parser',
                     expirationTime=timestamp,
                     data=pickle.dumps(self.schedule))

    def get_workflow_tokens(self):
        """Create Pinball tokens representing a workflow instance.

        Convert workflow jobs to tokens and create event tokens in inputs of
        top-level jobs.

        Returns:
            A list of job and event tokens representing a workflow instance.
        """
        all_jobs = self._get_transitive_deps()
        instance = get_unique_workflow_instance()
        result = []
        for job in all_jobs:
            result.append(job.get_job_token(self.name, instance))
        top_level_jobs = self._get_top_level_jobs()
        for job in top_level_jobs:
            event = Event(creator='parser')
            event_name = Name(workflow=self.name,
                              instance=instance,
                              job=job.name,
                              input_name=Name.WORKFLOW_START_INPUT,
                              event='workflow_start_event')
            result.append(Token(name=event_name.get_event_token_name(),
                                data=pickle.dumps(event)))
        return result


class JobDef(object):
    """Job is defined by a template that translates to a config."""

    def __init__(self, job_name, job_template, workflow, job_params=None):
        self.name = job_name
        self.template = job_template
        self.workflow = workflow
        self.job_params = job_params
        self.inputs = []
        self.outputs = []
        # If the job does not have a specified priority, we default to 1
        if job_template.priority:
            self.priority = job_template.priority
        else:
            self.priority = 1.0
        self.dependents = None

    def __hash__(self):
        return self.get_qualified_name().__hash__()

    def __str__(self):
        deps_str = ','.join(dep.get_qualified_name() for dep in self.inputs)

        return ('job:%s, deps:[%s]' % (self.get_qualified_name(), deps_str))

    def __repr__(self):
        return self.__str__()

    def add_dep(self, job):
        self.inputs.append(job)
        job.outputs.append(self)

    def _get_dependents(self):
        if not self.dependents:
            self.dependents = set([self])
            for job in self.outputs:
                self.dependents = self.dependents.union(job._get_dependents())
        return self.dependents

    def compute_score(self):
        """Each node's score is computed as the sum of the priorities of each
        of the nodes in the node's sub-tree"""
        dependents = self._get_dependents()
        return sum((dependent.priority for dependent in dependents))

    def get_qualified_name(self):
        """Get fully qualified job name that includes the workflow name."""
        return _get_qualified_name(self.workflow.name, self.name)

    def get_canonical_name(self, workflow_name):
        """Get a unique name identifying a job in the context of a workflow.

        Canonical names are required in addition to qualified names because
        the same job may be part of multiple workflows (due to cross-workflow
        job dependencies). Azkaban requires jobs to have globally unique
        names.
        """
        _CANONICAL_DELIMITER = '_'
        canonical_name = self.get_qualified_name().replace(
            _NAME_DELIMITER, _CANONICAL_DELIMITER)
        if workflow_name == self.workflow.name:
            return canonical_name
        return (workflow_name + _CANONICAL_DELIMITER + canonical_name)

    def _get_template_params(self):
        """Get params that will be passed down to job template.
        """
        return self.job_params

    def get_job_token(self, workflow_name, workflow_instance):
        """Convert the job to a pinball token representing its properties.
        Condition is similar to job. We check the template to decide use
        job_templates or condition_templates dynamically.

        Args:
            workflow_name: The name of the workflow in which context this job
                is instantiated.
            workflow_instance: The workflow instance of the output job token.

        Returns:
            A pinball token representing the job.
        """

        if self.workflow.name == workflow_name:
            inputs = [input_job.name for input_job in self.inputs]
        else:
            # If it's an external job, do not include its inputs.
            inputs = []
        if not inputs:
            inputs = [Name.WORKFLOW_START_INPUT]
        outputs = []
        for output_job in self.outputs:
            if output_job.workflow.name == workflow_name:
                outputs.append(output_job.name)

        if issubclass(self.template.__class__, JobTemplateBase):
            params = self._get_template_params()
            job = self.template.get_pinball_job(inputs, outputs, params)
            name = Name(workflow=workflow_name,
                        instance=workflow_instance,
                        job_state=Name.WAITING_STATE,
                        job=self.name)
            result = Token(name=name.get_job_token_name(), data=pickle.dumps(job))
            result.priority = self.compute_score()
        elif issubclass(self.template.__class__, ConditionTemplateBase):
            condition = self.template.get_pinball_condition(outputs)
            name = Name(workflow=workflow_name,
                        instance=workflow_instance,
                        job_state=Name.WAITING_STATE,
                        job=self.name)
            result = Token(name=name.get_job_token_name(), data=pickle.dumps(condition))
        else:
            raise Exception("Template must be a subclass of JobTemplateBase or ConditionTemplateBase!")

        return result


# There is a cyclic import dependency between schedule and workflow.  Placing
# imports of such modules at the end of the file makes them work.
from pinball.scheduler.schedule import OverrunPolicy
from pinball.scheduler.schedule import WorkflowSchedule
